#=======================================================================================================
#PaREn Classifiers
#
#In this module you find functions for the application of different classifiers. In this module you cannot use
#PaREn Messages (dict) for reporting of errors and warnings. Therfore regular print is used.
#
#	Main Features:
#	
#	CombineNumericMessageFields	:	FUNCTION	Collect information from messages and put it into a single convenient ndarray
#	StatsOnNumericMessageField	:	FUNCTION	Evaluate min, max, mean, median, std and sum of the fields provided by the messages
# 	CombineTextualMessageFields	:	FUNCTION	Create a single text from many informational messages.
#	MergeMessages				:	FUNCTION	Merges totally different message types into one bigger message
#	CommonFields()				:	FUNCTION	Given a list of messages it finds the fields that are common to all messages.
#	RandomName()				:	FUNCTION	Generates a variable length random names from big latin letters
#	PrintMessage()				:	FUNCTION	Prints a PaREn Message on screen. PaREn Messages are dictionaries 
#												with a specific set of conventions. The key "Description" is
#												displayed as header of the message and does not have a ShadowKey.
#												Messages can contain ShadowKeys that have the same name with an
#												underscore: "ValueX" -> "_ValueX"
#												ShadowKeys are displayed differently. ShadowKeys are typically lists
#												that enumerate valid content or contain types that can be checked
#												against with a isinstance() command.
#	TerminalMessage()			:	FUNCTION	Prints a single message on screen. Used by PrintMessage.
#	HTMLMessage()				:	FUNCTION	Generates a HTML string that essentially does the same thing as
#												PrintMessage() in a browser.
#	WelcomeMessage()			:	FUNCTION	Generates a friendly welcom message
#	TypeOf()					:	FUNCTION	Given a message finds the type of the specified fieldname if shadow fields are provided. If no shadow field is provided TypeOf guesses from what it sees on the field.
#	Filter()					:	FUNCTION	Filter piles of messages for messages with certain combinations of fields (AND/OR) and
#												specific conditions of the fields (expressible in python syntax supplied as a string)
#	FixPhysicalProperties()		:	FUNCTION	Helper. Checks ndarray type objects to be of specified number of physical dimensions: 1=[], 2 = [[]] ...
#	Fix2DLogicalProperties()	:	FUNCTION	Helper. Does some translating, resizing of 2d ndarray to match required width and height
#	Graphs()					: 	FUNCTION	Creates a logical graph that can be displayed using PaREnGUI.DrawGraph()
#	NoneField()					:	FUNCTION	Creates a list of None elements of variable size.
#	Answered()					:	FUNCTION	Helps to evaluate TerminalInteraction() if you want to do this yourself
#	MatchesWildcard()			:	FUNCTION	Performs a simple string matching based on * and ? wildcards as you know from the file system.
#	RenameFields()				:	FUNCTION	Rename a set of fields to have a common name. This can be usefull, when system
#												components return the same 'thing' under a slightly different caption. You can use
#												wildcards to specify fieldnames and apply the operation to a pile of messages at once.
#	DeleteFields()				:	FUNCTION	Works like RenameFields() but it deletes the specified fields.
#	AddFields()					:	FUNCTION	Adds fields to listed messages.
#	Select()					:	FUNCTION	Filter piles of messages for messages with certain combinations of fields (AND/OR) and
#												specific conditions of the fields (expressible in python syntax supplied as a string)
#	UniqueMessages()			:	FUNCTION	Removes duplicates 
#	UniqueStates()				:	FUNCTION	Returns unique values found for a given fieldname in a supplied message set.
#	IsNumeric()					: 	FUNCTION	Looks if expression is numeric
#	NewMessage()				: 	FUNCTION	Creates a message stub
#	HTMLPage()					:	FUNCTION	Creates an HTML page from a list of messages
#	SetDescription()			:	FUNCTION	Sets Description attribute in messages
#	SetStatus()					:	FUNCTION	Sets Status attribute in message
#	AssertVector()				:   FUNCTION	Converts a vector in a shape of your liking
#	AssertWorkingDirectory()	: 	FUNCTION	Changes your working directory to the specified path. If path does not exist, it will be created.
#	AverageKurtosis()			:	FUNCTION	Average Kurtosis from Kurtosis values for each dimension
#	Kurtosis()					:   FUNCTION	Kurtosis for each of the dimensions in the data.
#	IsVector()					:	FUNCTION	Is the parameter a vector?
#	IsRectMatrix()				:	FUNCTION	Is parameter a rectangular matrix?
#=======================================================================================================


#=======================================================================================================
#INCLUDES
#=======================================================================================================
from numpy import *;
from scipy import *;
from pylab import *;

import os;
import sys;
import time;
import string;


#=======================================================================================================
#DEFAULTS
#=======================================================================================================
__author__      			= "Aleksander Lodwich"
__copyright__  				= "Copyright (C) 2008 DFKI"
__version__     			= "0.1"

ModuleVerbosity				= 1;
PrintingToTerminalAllowed	= 1;


#=======================================================================================================
#FUNCTION AddFields()
#RETURNS LIST OF DICTIONARIES
#Adds a specific set of key:value pairs to the list of messages
#Supply				:	Dictionary (a PaREn Message) that will be added to the messages
#ListOfMessages		:	LIST OF DICTIONARIES. The messages you would like to 'compact'
#=======================================================================================================
def AddFields( Supply, ListOfMessages ):
	for Message in ListOfMessages:
		Message.update( Supply );
	return ListOfMessages;


#=======================================================================================================
#METHOD AddVertical( *Matrices )
#RETURNS 2D Array
#This method is a concatenate() alias that concatenates matrices vertically:
#Example:
#    x = rand( 5, 3 );
#    y = rand( 12, 3 );
#    z = AddVertical( x, y ); -> z is 17x3
#Matrices	: 2D Array or None
#=======================================================================================================
def AddVertical( *Matrices ):
	M = [];
	for Matrix in Matrices:
		if Matrix is not None:
			M += [Matrix];
	return concatenate( M, 0 );

#=======================================================================================================
#METHOD AddHorizontal( *Matrices )
#RETURNS 2D Array
#This method is a concatenate() alias that concatenates matrices vertically:
#Example:
#    x = rand( 5, 3 );
#    y = rand( 5, 17 );
#    z = AddHorizontal( x, y ); -> z is 5x20
#Matrices	: 2D Array or none
#=======================================================================================================
def AddHorizontal( *Matrices ):
	M = [];
	for Matrix in Matrices:
		if Matrix is not None:
			M += [Matrix];
	return concatenate( M, 1 );




#=======================================================================================================
#FUNCTION Answered()
#RETURNS STRING
#Answer			:	STRING. The string that you expect as answer
#Message		:	DICTIONARY. PaREn Message that was in TerminalInteraction()
#=======================================================================================================
def Answered( Answer, Message ):
	if type( Message ) != dict:
		return '';
	
	if Answer.upper() == Message.get( 'ANSWER', '' ):
		return 1;
	return 0;


#=======================================================================================================
#FUNCTION AssertVector
#RETURNS NDARRAY
#Converts into a numpy vector of your liking. 
#MaybeVector	:	NDARRAY? LIST? TUPLE? It should be at least convertible into a numpy array and the data should look like a vector.
#PhysicalDims	:	INTEGER. >0. Number of dimensions to convert your vector into. More than 2 don't actually make sense
#Direction 		:	INTEGER. 0: vertical 1:horizontal 2.? 3...
#=======================================================================================================
def AssertVector( MaybeVector, PhysicalDims=2, Direction=0 ):
	if IsVector( MaybeVector ):
		if Direction < 0 or Direction >= PhysicalDims:
			print "PaREnUtility::AssertVector(): Direction must be: 0<=Direction<PhysicalDims. Parameters were: Direction=", Direction, " PhysicalDims=", PhysicalDims
			return array([]);
			
		Vector = FixPhysicalProperties( MaybeVector, PhysicalDims, "PaREnUtility::AssertIsVector()", 0 );
		Length = max(shape( Vector ));
		Shape  = ones( PhysicalDims );
		Shape[Direction] = Length;
		Vector = Vector.reshape( Shape );
		return Vector;
	else:
		try:			
			Vector = array( MaybeVector );
			if Vector.ndim > 2:
				print "PaREnUtility::AssertVector(): Parameter is not a vector! Parameter dimensionality was: ", Vector.ndim;
				return array([]);				
			return AssertVector( Vector, PhysicalDims, Direction );
		except:
			print "PaREnUtility::AssertVector(): Parameter is not convertible to numpy array. Parameter type was: ", type( MaybeVector );
			return array([]);
	#end if
#end def AssertVector


#=======================================================================================================
#FUNCTION AssertWorkingDirectory
#RETURNS String
#Essentially, this is a 'cd' command but it will also create the path if non existent.
#It returns the previous path, so you can revert to it later.
#Path		: STRING. Windows or Unix path
#=======================================================================================================
def AssertWorkingDirectory( Path ):
	#first split the paths
	CurDir = os.getcwd();
	
	if Path is None:
		return CurDir;
	
	Windows = Path.split( '\\' );
	Unix	= Path.split( '/' );
	
	#the split with more parts is correct:
	Elements = [];
	if len( Unix ) > len( Windows ):
		Elements = Unix;
		Separator = '/';
	else:
		Elements = Windows;
		Separator = '\\';
		
	#now try all depths of path in order to see, whether they exist. If not, create
	for i in range( 1, len( Elements ) ):
		ConstructedPath = Elements[0];
		
		for j in range( 1, i+1 ):
			ConstructedPath = ConstructedPath + Separator + Elements[j];

		if not os.access( ConstructedPath, os.F_OK ):
			os.mkdir( ConstructedPath );
	#end for
	
	#finally, go to that place.
	os.chdir( Path );
	return CurDir;
#end def 


#=======================================================================================================
#FUNCTION AssureStringCompatibility( Name, AllowedCharacters, SubstituteCharacter)
#RETURNS STRING
#Replaces illegal characters with a legal character.
#Name				: STRING. Name to be checked
#AllowedCharacters	: STRING. String with letters allowed to be in name;
#SubstituteCharacter: STRING. Specify a single (multiple) characters to be placed where bad characters occured.
#=======================================================================================================
def AssureStringCompatibility( DangerousName, AllowedCharacters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-", SubstituteCharacter = "_" ):
	SafeName = "";
	for Character in DangerousName:
		if Character in AllowedCharacters:
			SafeName += Character;
		else:
			SafeName += SubstituteCharacter;
	return SafeName;


#=======================================================================================================
#FUNCTION AverageKurtosis()
#RETURNS FLOAT
#Calculates a single kurtosis value for multidim data.
#data		: 2d ndarray. Rows are vectors. Columns are feature dimensions.
#=======================================================================================================
def AverageKurtosis( data ):
	return mean( Kurtosis( data ) );


#=======================================================================================================
#FUNCTION CombineNumericMessageFields
#RETURNS Dictionary
#Calcualtes Min, Max, Sum, Mean, Median, Average on a given numerical Field from the messages supplied
#Fields can be either scalar, 1D nd arrays or 1D ndlists
#Fieldname		:	STRING. Exact name of field. Can also be a list of names.
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def CombineNumericMessageFields( Fieldname, Messages ):
	#SECTION A: GUARD
	if type( Fieldname ) == list:							#Sometimes the user wants the action repeated for many fields
		Fieldnames = Fieldname;
	else:
		Fieldnames = [Fieldname];
	#end if;
	FinalResult = {};
	
	#SECTION B: SEARCH FOR MINIMUM WIDTH OF DATA
	Width 	= inf;											#Barely ever the merged numeric fields have the same number of elements
	for Message in Messages:								#this is indeed not a problem because the same type of processes
		Entry = Message.get( Fieldname, None );				#might get recorded longer or shorter.
		if Entry is None:									#if the fields are of the same length it would not harm, of course!
			continue;
		if isscalar( Entry ):
			Width = 1;
		elif len( Entry ) < Width:
			Width = len( Entry );
		#end if
	#end for

	#SECTION C: ACTUALLY PERFORM MERGING
	for Fieldname in Fieldnames:
		Result	= { 'Description':'Combined data of "' + Fieldname + '" retrieved from ' + str( len( Messages ) ) + ' messages.', 'Status':'OK' };
		Data 	= [];
		N		= 0;
		#SECTION C.1: PUT DATA INTO 2D LISTS
		for Message in Messages:
			if type( Message ) != dict:
				if ModuleVerbosity:
					print "PaREnUtility::CombineNumericMessageFields(): Error! Message must be of type DICTIONARY!";
				continue;

			Entry = Message.get( Fieldname, None );
			if Entry is None:
				continue;
			if isscalar( Entry ):
				Data += [[Entry]];
			elif Width < inf:
				Data += [Entry[:Width]];
			N += 1;
		#end for
		
		try:
			Data			= FixPhysicalProperties( array( Data ), 2, 'PaREnUtility::CombineNumericMessageFields()', 1 );
		except:
			Result['Status'] = 'Data is inconsistent in the messages!';
			return Result;
		#end try
		
		Result['Data']	= Data;
		Result['N']		= N;
		FinalResult = MergeMessages( [ FinalResult, Result ] );
	#end for
	return FinalResult;
#end def


#=======================================================================================================
#FUNCTION CombineTextualMessageFields
#RETURNS Dictionary
#Calcualtes Min, Max, Sum, Mean, Median, Average on a given numerical Field from the messages supplied
#Fields can be either scalar, 1D nd arrays or 1D ndlists
#Fieldname		:	STRING. Exact name of field. Can also be a list of names;
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def CombineTextualMessageFields( Fieldname, Messages ):
	#SECTION A: GUARD
	if type( Fieldname ) == list:					#Fieldname is sometimes supplied as list meaning that the user wants to repeat the action for all the fields.
		Fieldnames = Fieldname;
	else:
		Fieldnames = [Fieldname];
	#end if;
	FinalResult = {};

	#SECTION B: APPLY FIELD MERGING
	for Fieldname in Fieldnames:
		Result	= { 'Description':'Concatenated data from "' + Fieldname + '" retrieved from ' + str( len( Messages ) ) + ' messages.', 'Status':'OK' };
		Data 	= '';
		N		= 0;
		for Message in Messages:
			Entry = Message.get( Fieldname, None );
			if Entry is None:
				continue;
			if type( Entry ) != str:
				Entry = str( Entry );
			Data += Entry + ' ';
			N 	 += 1;
		#end for
		
		Result['Text'] 	= Data;
		Result['N']		= N;
		FinalResult = MergeMessages( [ FinalResult, Result ] );
	#end for
	
	#SECTION C: PACKAGING
	return FinalResult;
#end def


#=======================================================================================================
#FUNCTION CommonFields()
#RETURNS LIST of STRINGS
#Given a list of messages it finds the fields that are common to all messages.
#Messages		:	LIST. List of PaREn Messages that shall be renamed.
#=======================================================================================================
def CommonFields( Messages ):
	if type( Messages ) is not list and type( Messages ) is not tuple:
		return [];
	
	N = len( Messages );
	Counter = {};
	for Message in Messages:
		Keys = Message.keys();
		for Key in Keys:
			Counter[Key] = Counter.get( Key, 0 ) + 1;

	Result = [];
	Pairs = Counter.items();
	for Pair in Pairs:
		if Pair[1] == N:
			Result += [Pair[0]];
	
	return Result;
#end def CommonFields()



#=======================================================================================================
#FUNCTION DataRadius()
#RETURNS FLOAT
#Calculates how compact the data set is.
#data			:	2D NDARRAY. data where rows are vectors
#=======================================================================================================
def DataCompactness( data ):
	S = DataRadii( data );
	return std( S ) / max( S - mean( S ) );


#=======================================================================================================
#FUNCTION DataRadii()
#RETURNS LIST OF FLOAT
#Calculates all distances between all vectors. This process can be limited using MaxVal subsamples.
#This function can be used to get representative relationships between blobs in feature space.
#data			:	2D NDARRAY. data where rows are vectors
#MaxVal			:	INT > 1. Maximum number of vectors to use. 
#=======================================================================================================
def DataRadii( data, MaxVal=100 ):
	Data = data;
	N = VectorCount( data );
	if N > MaxVal:
		Data = data[random.permutation( range( N ) )[:MaxVal]];
		N = MaxVal;
		
	S = [];
	for i in range( N ):
		for j in range( i+1, N ):
			S += [norm( Data[i] - Data[j] )];
	
	return S;


#=======================================================================================================
#FUNCTION DataRadius()
#RETURNS FLOAT
#Calculates the radius of the data world.
#data			:	2D NDARRAY. data where rows are vectors
#=======================================================================================================
def DataRadius( data ):
	CenterOfData 	= mean( data, 0 );		
	Radius 			= 0;
	Radii 			= [];
	
	for row in data:
		Radii += [norm( row - CenterOfData )];
	
	Radii 		= array( Radii );
	Radius 		= Radii.max();
	return Radius;


#=======================================================================================================
#FUNCTION DataFilter()
#RETURNS 1D NDARRAY
#Performs filtering using a supplied function on "time series". 
#data			:	1D NDARRAY or LIST. "Time Series"
#func			:	FUNCTION HANDLE. Functions of the type like median(), mean(), diff(), std() ...
#windowsize		:	INT > 1.	Number of samples to take into function.
#step			:	INT > 1 or -1 WHEN -1 step will have size of window. Other common values are i.e. 1.
#=======================================================================================================
def DataFilter( data, func, windowsize, step=-1 ):
	Result = [];
	Mode = 0;
	try:
		func( [1,2,3], 0 );
	except:
		Mode = 1;
		
	if step == -1:
		step = windowsize;
	
	if Mode == 0:
		for i in range( 0, VectorCount( data ), step ):
			Result += [func(data[i:i+windowsize], 0 )];
	else:
		for i in range( 0, VectorCount( data ), step ):
			Result += [func(data[i:i+windowsize] )];
	return array( Result );


#=======================================================================================================
#FUNCTION DeleteFields()
#RETURNS FLOAT, STRING
#This deletes message fields with specified field names.
#FieldNameList	:	LIST/TUPLE.	List of string representing fieldnames. Fieldnames can be augmented with * and ? wildcards. Use \* or \? in order to specify regular name elements.
#Messages		:	LIST. List of PaREn Messages that shall be renamed.
#CaseSensitive	:	1/0. If set to 1 then name comaparisons are done in a case sensitive manner.
#=======================================================================================================
def DeleteFields( FieldNamesList, Messages, CaseSensitive=0 ):
	#SECTION A: INITIALIZATION
	Comments 	= 'OK';
	Code		= 1;
	
	#SECTION A: GUARD
	if type( FieldNamesList ) == str:
		FieldNamesList = [FieldNamesList];
	
	if type( FieldNamesList ) != list and type( FieldNamesList ) != tuple:
		return 0, "Wrong type of data for FieldNamesList!";

	if type( NewName ) != str:
		return 0, "Wrong type of data for NewName: Must be a string!";

	#SECTION B: RENAMING
	for Message in Messages:
		Keys = Message.keys();
		for OldName in FieldNamesList:
			for Key in Keys:
				if MatchesWildcard( OldName, Key, CaseSensitive ):
					Message.pop( Key );
				#end if MatchesWildcard
			#end for all keys
		#end for all old names
	#end for all messages
	return Code, Comments
#end def



def ExtractData( InFields, OutFields, Messages, Filename=None, WriteStats=1, WriteRaw=1 ):
	#SECTION INDEX:
	#
	#SECTION A: GUARD
	#SECTION B: INITIALITZATION
	#SECTION C: GATHER INFORMATION AND PUT IT INTO A STRUCTURE AllResults
	#SECTION D: USE THE STRUCTURE AllResults IN ORDER TO WRITE A FILE
	#SECTION E: CLEAN UP
	
	
	#SECTION A: GUARD
	if type( InFields ) is str:
		InFields = [InFields];

	if type( OutFields ) is str:
		OutFields = [OutFields];

	if type( Messages ) is dict:
		Messages = [Messages];
	
	if type( InFields ) is not list and type( InFields ) is not tuple:
		return [];
	
	if type( OutFields ) is not list and type( OutFields ) is not tuple:
		return [];

	if type( Messages ) is not list and type( Messages ) is not tuple:
		return [];

	WriteStats 	= InRange( 0, WriteStats, 1 );
	WriteRaw 	= InRange( 0, WriteRaw,   1 );

	#SECTION B: INITIALITZATION
	AllResults = [];
	MaxRange = 0;
	
	#SECTION C: GATHER INFORMATION AND PUT IT INTO A STRUCTURE AllResults
	for OutField in OutFields:
		Result = { "Output": OutField };
		
		ValueRanges = {};
		for InField in InFields:
			Range = UniqueStates( InField, Messages );
			ValueRanges[InField] = Range;
			
		Result['ValueRanges'] = ValueRanges;
		Combinations = Select( OutFields, Messages, InFields ); #take messages that have OutFields and group them by InFields
		
		ContentList = [];
		for MessagesForCombination in Combinations:
			Entry = {};
			for InField in InFields:
				Value = UniqueStates( InField, MessagesForCombination )[0]; #there is always exactly one. Therfore we can do [0]
				Entry[InField] = Value;
				
			Entry['Data'] = CombineNumericMessageFields( OutField, MessagesForCombination )['Data'];
			ContentList += [Entry];
			MaxRange = max( MaxRange, len( Entry['Data'] ) );
		Result['Content'] = ContentList
		AllResults += [Result];
		
	#Exit here if writing to file is not neccessary:
	if Filename is None or ( not WriteStats and not WriteRaw ):
		return AllResults;
	
	#SECTION D: USE THE STRUCTURE AllResults IN ORDER TO WRITE A FILE
	OutputFile = open( Filename, 'w' );
	for Result in AllResults:
		#Write title and description:
		OutputFile.write( 'This section shows dependency of "' +  Result['Output'] + '" on the fields:' + "\n" );
		Ranges  = Result['ValueRanges'];
		Keys = Ranges.keys();
		for Key in Keys:
			OutputFile.write( str( Key ) + "\t" + str( Ranges[Key] ) + "\n" );
		
		#STATS
		if WriteStats:
			#Write column headers:
			OutputFile.write( "\n" + 'STATISTICS' + "\n" );
			Ranges  = Result['ValueRanges'];
			Keys = Ranges.keys();
			for Key in Keys:
				OutputFile.write( str( Key ) + "\t" );
			
			OutputFile.write( "Min\tMax\tMean\tMedian\tStdDev\n" );
			#Write data:
			for Entry in Result['Content']:
				Keys = Entry.keys();
				#at first write input parameters
				for Key in Keys: 
					if Key != 'Data':
						OutputFile.write( str( Entry[Key]) + "\t" );

				#then output data in the same line
				Data = Entry['Data'].T[0];
				OutputFile.write( str( Data.min() ) + "\t" + str( Data.max() ) + "\t" + str( Data.mean() ) + "\t" + str( median( Data ) ) + "\t" + str( Data.std() ) + "\n");
			OutputFile.write( "\n\n\n" );
		#end if WriteStats
		
		
		#RAW DATA
		if WriteRaw:

			SpreadSheet = [];
			MaxRowLength = 0;
			
			#Write column headers:
			Row = [];
			Ranges  = Result['ValueRanges'];
			Keys = Ranges.keys();
			for Key in Keys:
				Row += [ str( Key ) ];

			for i in range( MaxRange ):
				Row += [ 'D' + str(i)];				
			SpreadSheet += [Row];

			#Write data:
			for Entry in Result['Content']:
				Row = [];
				Keys = Entry.keys();
				#at first write input parameters
				for Key in Keys: 
					if Key != 'Data':
						Row += [str( Entry[Key])];
						
				#then output data in the same line
				Data = Entry['Data'].T[0];
				for Value in Data:
					Row += [str( Value ) ];
				SpreadSheet += [Row];
				MaxRowLength = max( MaxRowLength, len( Row ) );
			
		Columns = len( SpreadSheet );
		for Row in range( MaxRowLength ):
			for Col in range( Columns ):
				try:
					TextElement = SpreadSheet[Col][Row];
				except:
					TextElement = "";
				OutputFile.write( TextElement + "\t" );
			OutputFile.write( "\n" );
		OutputFile.write( "\n\n\n" );
		#end if WriteRaw
		
	#SECTION E: CLEAN UP
	OutputFile.close();
	return [];
#end def ExtractData()

#=======================================================================================================
#FUNCTION FieldIndex( Messages )
#RETURNS Dict, Dict
#This function identifies all fields available in a set of messages and assigns them unique id's.
#If you want just to know all the fields then call "AllFields = FieldIndex( Messages )[0].keys();"
#The first dictionary holds the relationship between fieldnames and id's and the second dictionary
#holds the relationship from id's to fieldnames. You can use id's for indexing in lists.
#Messages	: LIST of DICTs. Your database with messages. Could also be a PythonFileReader.
#=======================================================================================================
def FieldIndex( Messages ):	
	#for convenience support single message cases
	if type( Messages ) is dict:
		Messages = [Messages];

	try:
		Messages.__getitem__
	except:
		return {},{}
	#end try whether Messages is a list

	FieldsToPos = {};
	PosToFields = {};
	for Message in Messages:
		Keys = Message.keys();
		for Key in Keys:
			Pos = len( FieldsToPos );
			FieldsToPos[Key] = Pos;
			PosToFields[Pos] = Key;
		#end for
	#end for
	return FieldsToPos, PosToFields;
#end def 



#=======================================================================================================
#FUNCTION Filter()
#RETURNS LIST of LISTS (2D LIST) of DICTIONARIES (PaREn Messages)
#The function will return reduced messages containing only the specified fields. In opposite Select() will return all content of a message.
#FIELDS			: 	LIST OF STRINGS/STRING.	Fieldname(s) - case sensitive
#FROM			:	LIST OF DICTS. 			List of messages to scan
#GROUPBY		:	LIST OF STRINGS/STRING.	Fieldnames(s) to group by the results. - case sensitive
#WHERE			:	STRING.					Evaluable python string, where message attributes are available as variable names. Functions from sys, os and time modules can be used here. This can be a security risk!
#											(WHERE clause is a security risk! This is because statements to the system can be made using os.system() function.)
#
#USAGE EXAMPLE
#Messages	= [ {...}, {...}, {...} ];
#Selection	= Select( 'SomeField', Messages ); 						-> Returns only messages that have the field SomeField
#Selection	= Select( [ 'SomeField', 'OtherField' ], Messages ); 	-> Returns only messages that have the fields SomeField and OtherField
#Selection 	= Select( 'SomeField', Messages, WHERE='SomeField>2 and SomeField<5' );
#																	-> Returns only messages that have the field SomeField and this field has numeric values > 2 and < 5
#Selection	= Select( [ 'A', 'B' ], Messages, 'C' );				-> Messages with fields A,B and C where lists are grouped by values of C
#Selection	= Select( [ 'A', 'B' ], Messages, [ 'C', 'D' ], 'C+D/(A+int(B))<0.9' );
#																	-> returns messages that have fields A,B,C and D and that meet the condition 
#																		C+D/(A+int(B))<0.9 where B is a string type that must be converted to int.
#Selection 	= Select( FROM=Messages, GROUPBY='SomeField' );			-> returns messages that are grouped by the content of SomeField
#Selection 	= Select( FROM=Messages, GROUPBY='SomeField', WHERE='OtherField==1' );
#																	-> returns messages that are grouped by the content of SomeField. Note, that
#																	   all fields can be used for evaluation in the WHERE clause. When those
#																	   fields do not exist, the statement returns FALSE.
#ErroMessages = Select( 'Status', PaREnProtocol.ReportsOfModules(), WHERE='Status.upper()=="ERROR" and TIMESTAMP>time.time()-3600' );
#																	-> this would get you all error messages that occured within the last hour
#=======================================================================================================
def Filter( FIELDS=[], FROM=[], GROUPBY=[], WHERE=None ):
	Groups = Select( FIELDS, FROM, GROUPBY, WHERE );
	ResultMessages = [];
	for RelevantMessages in Groups:
		GroupMessages = [];
		for RelevantMessage in RelevantMessages:
			NewMessage = {};
			for Field in FIELDS:
				NewMessage[Field] = RelevantMessage[Field];
			GroupMessages += [NewMessage];
		ResultMessages += [GroupMessages];
	return ResultMessages;


#=======================================================================================================
#FUNCTION FixPhysicalProperties()
#RETURNS a Dims NumPy Array
#It checks and corrects physical dimensionality of ndarrays
#A:			Some ndarray object
#Dims:		int > 0 that specifies the no. of phys. dimensions for the ndarray
#Msg:		String specifying who requested the check
#Loud:		1/0 (Loud/Quiet)
#=======================================================================================================
def FixPhysicalProperties( A, Dims, Msg, Loud = 1 ):
	while len( shape( A ) ) < Dims:						#correct low ndarray dimensionality (physical)
		A = array( [ A ] );
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Correcting low ndarray dimensionality for matrix!";
		#end if
	
	while len( shape( A ) ) > Dims:						#correct high ndarray dimensionality (physical)
		A = A[0];
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Correcting high ndarray dimensionality for matrix!";
	#end while
	return A;


#=======================================================================================================
#FUNCTION Fix2DLogicalProperties()
#RETURNS a Dims NumPy Array
#It checks and corrects physical dimensionality of ndarrays
#A:			Some 2d ndarray object
#DimW:		int >= 0 that specifies the width of the 2d ndarray
#DimH:		int >= 0 that specifies the height of the 2d ndarray
#Msg:		String specifying who requested the check
#Loud:		1/0 (Loud/Quiet)
#=======================================================================================================
def Fix2DLogicalProperties( A, DimW, DimH, Msg, Loud = 1 ):
	ShapeOfA = shape( A );

	if DimW != DimH and ShapeOfA[0] == DimW and ShapeOfA[1] == DimH:
		A = transpose( A );							#this is probably just a transposition error
		return A;
	
	MaxDim = max( DimW, DimH );
	
	if shape( A )[1] > DimW:						#catch logical dimensions mismatch
		A = A[:,0:DimW];
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Removing logical dimensions from matrix!";
	#end if
	
	if shape( A )[1] < DimW:						#catch logical dimensions mismatch
		Reference = eye( MaxDim );
		A = concatenate( (A, Reference[:shape( A )[0], shape( A )[1]:DimW]), 1 );
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Adding logical dimensions to matrix!";
		#end if
	#end if

	if shape( A )[0] > DimH:						#catch logical dimensions mismatch
		A = A[0:DimH,:];
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Removing logical dimensions from matrix!";
	#end if
	
	if shape( A )[0] < DimH:						#catch logical dimensions mismatch
		Reference = eye( MaxDim );
		A = concatenate( (A, Reference[shape( A )[0]:DimH, :shape( A )[1]]), 0 );
		if ModuleVerbosity and Loud:
			print Msg + " - ERROR: Adding logical dimensions to matrix!";
		#end if
	#end if
	
	return A;
#end def Fix2DLogicalProperties()


#=======================================================================================================
#FUNCTION FormatFreeCSVTable( Filename, Data )
#RETURNS VOID
#Writes a comma separated file from data. Data is either a list or a list of lists. If Data is a list
#then each entry will be written into own row. If data is consisting of other items that support
#the list interface, then it will write strings for every embedded item. Don't supply objects
#that don't support __str__ or return meaningless content from it. This function doesn't watch you
#and if you need always the same number of entries in each row, you must provide them by putting
#None values into the right place in Data.
#
#Filename	: STRING. Path to the CSV file
#Data		: LIST. Each entry is a row. If row is a list, then its data is columns.
#=======================================================================================================
def FormatFreeCSVTable( Filename, Data ):
	#Here checking that Data supports reading list operation
	try:
		Data.__getitem__
	except:
		return;
	
	CSVFile = open( Filename, 'w' );
	for Row in Data:	#no type checking, that Data is a list: can be anything that supports __getitem__
		if Row is None:	#empty rows are ignored
			continue;
		try: #if rows is kind of a list, then write out its values separated by a semicolon
			Row.__getitem__ #if has list interface
			for Item in Row:
				if Item is None:
					CSVFile.write( ";" );
				else:
					try:
						CSVFile.write( str( Item ).replace( "\n", " " )+";" );
					except:
						CSVFile.write( 'ERROR: Object does not support serialization!;' );					
					#end try object supports __str__
				#end if item is none
			#end for item im row
		except: #if rows is not a king of a list, then print it directly
			try:
				CSVFile.write( str( Row ).replace( "\n", " " ) );
			except:
				CSVFile.write( 'ERROR: Object does not support serialization!;' );
			#end try object supports __str__
		#end try Row is a list
		CSVFile.write( "\n" );
	#end for Row in Data
	
	CSVFile.close();
#end def FormatFreeCSVTable()


#=======================================================================================================
#FUNCTION Graphs()
#RETURNS a LIST OF DICTIONARIES
#Creates logical graphs for a specific input field and sepcific output field. Messages are 
#a list of protocolized messages during an experiment.
#InField		: STRING. Name of parameter field. x in y=f(x)
#OutFields		: STRING OR LIST OF STRINGS. Name of function. y in y=f(x)
#Messages		: LIST OF DICTIONARIES. Messages obtained during experiment.
#=======================================================================================================
def Graphs( InField, OutFields, Messages ):
	if type( OutFields ) is not list:
		OutFields = [OutFields];
		
	ResultGraphs = [];
	for OutField in OutFields:
		Graph = NewMessage( 'Graph for ' + OutField + ' given a ' + InField );
		Graph['Function'] = OutField;
		Graph['Parameter'] = InField;
		Values = UniqueStates( InField, Messages );
		Graph['ParameterValues'] = Values;
		
		TempMax = [];
		TempMin	= [];
		TempMed = [];
		TempAvg	= [];
		TempStd	= [];
		
		for Value in Values:
			if IsNumeric( Value ):
				WhereClause = (InField+"=="+str(Value));
			else:
				WhereClause = (InField+"=='"+str(Value)+"'");
			Selection = Select( OutField , Messages, WHERE=WhereClause  )[0];
			Stats = StatsOnNumericMessageField( OutField, Selection );
			TempMax += [Stats['Max']];
			TempMin += [Stats['Min']];
			TempMed += [Stats['Median']];
			TempAvg += [Stats['Avg']];
			TempStd	+= [Stats['Std']];
			
		Graph['Min'] = array(TempMin).T[0];
		Graph['Max'] = array(TempMax).T[0];
		Graph['Avg'] = array(TempAvg).T[0];
		Graph['Median'] = array(TempMed).T[0];
		Graph['Std'] = array(TempStd).T[0];
		ResultGraphs += [Graph];
		
	return ResultGraphs;
#end def Graphs();


#=======================================================================================================
#FUNCTION GroupedDataRadii()
#RETURNS DICTIONARY
#Calculates for every point in the data the distances to the other points. The key to the radii of every
#point is its index. STRING Indices with preceeding underscore indicate target vector.
#Because this function is quite expensive the MaxVal limits the number of representants.
#data			:	2D NDARRAY. Data where rows are vectors
#MaxVal			:	INT > 1. Maximum number of samples picked from data.
#=======================================================================================================
def GroupedDataRadii( data, MaxVal=100 ):
	Data = data;
	N = VectorCount( data );
	Indices = random.permutation( range( N ) )[:MaxVal];

	if N > MaxVal:
		Data = data[Indices];
		N = MaxVal;
		
	AllRadii = {};
	for i in range( N ):
		Radii = [];
		for j in range( N ):
			if i != j:
				Radii += [norm( Data[i] - Data[j] )];
		AllRadii[Indices[i]] = Radii;
		AllRadii["_"+str(Indices[i])] = j;
		
	return AllRadii;


#=======================================================================================================
#FUNCTION HTMLMessage
#RETURNS STRING
#HTML translation of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#=======================================================================================================
def HTMLMessage( Message, Loud = 1, PadLines = 0, FullView = 0  ):
	#SECTION A: GUARDS SECTION
	if not PrintingToTerminalAllowed:
		return;
	
	if not Loud:
		return;
	
	if not isinstance( Message, dict ):
		return;
	
	if len( Message.keys() ) == 0:
		return;
	
	#SECTION B:	Print Header
	Output = "";
	ShouldPrintDataBar = 1;
	if Message.get( "Description", None ) is not None:
		Output += "<h2>Description</h2>\n"
		Output += Message["Description"];										#The following stuff is just a complicated print.
		ShouldPrintDataBar = 1;
	#end if

	#SECTION C: CHECK FOR INVISIBLE FIELDS
	OptimalBorder = 0;
	Keys = Message.keys();														#Get all the keys we would like to present
	
	Remarks = '';
	for i in range( len(Keys)-1,-1,-1 ):
		if type( Keys[i] ) != str:			
			Remarks += str( Keys.pop(i) ) + ' ';
		#end if
	#end for
	
	if len( Remarks ) > 0:
		Message = Message.copy();
		Message['Remarks'] = 'The following fields can not be displayed:' + Remarks;
		Keys += ['Remarks'];
	#end if

	#SECTION D: WRITE DATA SECTION
	HasPrintedDataBar 	= 0;
	StatusKey 			= None;
	LengthLimit			= 60;
	for key in Keys:
		if key.upper() == "DESCRIPTION":										#This key has been printed already
			continue;

		if key.upper() == "STATUS":												#We will deal with this later.
			StatusKey = key;
			continue;
		
		if key[0] == '_':														#If a key is not a ShadowKey then print it like that:
			continue;

		if ShouldPrintDataBar and not HasPrintedDataBar:
			Output += "<h2>Data</h2>\n<table>\n";
			HasPrintedDataBar = 1;

		Value = Message[key];													#Find the value of the key
		if Value is None:														#Sometimes Values are None to indicate a Default setting
			Value = "Default";
			
		if not FullView and type( Value ) == dict and len( str( Value ) ) > LengthLimit:
			Value = "(PaREn Message)";
		
		
		if not FullView and len( str( Value ) ) > LengthLimit:					#In this clause we shorten up the look of the strings
			NewValue = str(Value)[:LengthLimit] + '...';						#and add some extra infos in compensation
			if type( Value ) == list:
				NewValue += "\n" + "LIST of " + str( len( Value ) ) + " elements.";
			if type( Value ) == tuple:
				NewValue += "\n" + "TUPLE of " + str( len( Value ) ) + " entries.";
			if type( Value ) == ndarray:
				NewValue += "\n" + "ARRAY of " + str( shape( Value ) ) + " size.";
			Value = NewValue;
		
		if len( shape( Value ) ) > 1 and type( Value ) != str:					#this clause deals with 2d arrays that need some more space.
			Value = '\n' + str( Value ) + '\n';			
			Value = str(Value).replace( '\n', '\n|'+string.rjust( "", OptimalBorder + 6 ) );
		elif str( Value ).find( '\n' )+1:
			Value = str(Value).replace( '\n', '\n|'+string.rjust( "", OptimalBorder + 6 ) );			
		#end if
		
		if Message.get( '_' + key, None ) is not None:						#if it has as shadow key then print...
			ListMotivatedInsertion = "\t"
			Type = Message['_'+key]

			if type( Type ) == list or type( Type ) == tuple:
				ListMotivatedInsertion = " where possible are: ";

			Output += "<tr><td>" + key + "</td><td>" + str( Value )+ListMotivatedInsertion + str( Message['_'+key] ) + "</td><tr>\n"; #...it with the legal Range
		else:
			Output += "<tr><td>" + key + "</td><td>" + str( Value ) + "</td><tr>\n";				#otherwise just print it.
		#end if
	#end for	
	
	if HasPrintedDataBar:
		Output += "</table>\n";
	
	#SECTION D: Print Status
	if Message.get( StatusKey, None ) is not None:
		Value = Message[StatusKey];
		Output += "<div align=center> <h2>Status</h2></div>\n" ;
		if type( Value ) == str:
			if str(Value).upper().find( "OK" )+1 or str(Value).upper().find( "SUCCESS" )+1\
			or str(Value).upper().find( "ALRIGHT" )+1 or str(Value).upper().find( "GOOD" )+1:
				Output += "<div align=center> >> " + Value + " << </div>\n";			
			elif str(Value).upper().find( "WARNING" )+1:
				Output += "<div align=center> * * * " + Value + " * * * </div>\n";	
			elif str(Value).upper().find( "ERROR" )+1 or str(Value).upper().find( "FAIL" )+1:
				Output += "<div align=center> :( " + Value + " :( </div>\n";
			else:
				Output += "<div align=center>" + Value + "</div>\n";
			#end if
		#end if TYPE IS STRING
			
		if type( Value ) == type( 0.0 ):
			f = max( min( Value, 1.0 ), 0.0 );
			c = fix( 60 * f );
			Bars = "";
			for i in range( int( c ) ):
				Bars += "|";
			Output +=  "<div align=center>" + ( "[" + string.ljust( Bars, 60 ) + "]" + str(fix(100*f)) + "%").replace( ' ', '.' ) + "</div>\n";
		#end if TYPE IS FLOAT

		if type( Value ) == type( 0 ):
			f = max( min( Value, 100 ), 0 );
			c = fix( f * 0.6 );
			Bars = "";
			for i in range( int( c ) ):
				Bars += "|";
			Output += "<div align=center>" + ("[" + string.ljust( Bars, 60 ) + "]" + str(f) + "%").replace( ' ', '.' ) + "</div>\n";
		#end if TYPE IS INT
	#end if
	Output += "<hr />\n";
	
	#SECTION E: Write Pad Lines
	for i in range( PadLines ):
		Output += "<br>";
		
	return Output;
#end def HTMLMessage


#=======================================================================================================
#FUNCTION Kurtosis()
#RETURNS LIST of FLOATS
#Calculates kurtosis for each dimension
#data		: 2d ndarray. Rows are vectors. Columns are feature dimensions.
#=======================================================================================================
def Kurtosis( data ):
	D = data.copy();
	m = mean( D );
	D -= m;
	M4 = D**4;
	M2 = D**2;
	N = VectorCount( data );
	Result = ( N * sum( M4, 0 ) / sum( M2, 0 )**2 ) - 3;
	return Result;


#=======================================================================================================
#FUNCTION PrintHTML
#RETURNS VOID
#Pretty print of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#Filename	:   STRING. 	HTML file to be written into
#Refresh	:	INTEGER.	Number of seconds, before html is reloaded
#PadLines	:	Integer. 	Prints PadLines empty lines after the message
#FullView	:	1/0.		Abbreviate lists and arrays?
#=======================================================================================================
def HTMLPage( Messages, Title='PaREn', Refresh = 0, PadLines = 0, FullView = 0 ):
	if type( Messages ) == dict:
		Messages = [Messages];

	Result = "";
	if type( Messages ) == list or type( Messages ) == tuple:
		
		Result += "<head><title>"+str(Title)+"</title>"
		if Refresh>0:
			Result += "<META HTTP-EQUIV=REFRESH CONTENT=" + str( Refresh ) + ">"
		Result += "</head>\n<body><div align=center><h1>"+str(Title)+"</h1></div><hr />\n";
		
		for Message in Messages:
			Result += HTMLMessage( Message, 1, PadLines, FullView );
		#end for
	#end if
	return Result + "</body>";
#end def 


#=======================================================================================================
#FUNCTION HTMLTable( Filename, Data, InterestingFields, FieldColumn, FieldRows, InfoTypes )
#RETURNS 1/0 (SUCCESS / FAILED )
#This functions takes a list (or other list like objects like PythonFileReader) of messages
#and extract tabular information from the messages concerning >InterestingFields<. The
#number of rows and columns is automatically decided, depending how many different values
#are available in the database. This function only works for numerical types of >InterestingFields<.
#You can choose to generate statistical tables for average, median, standard deviation, maximum and minimum.
#
#Filename			:	STRING. Path to a new or overwritable html file
#Data				: 	LIST of DICTS. Database with messages.
#InterestingFields	:	STRING or LIST of STRINGS. Specifies which numerical field(s) you're interested in.
#FieldColumn		:	STRING. Specifies exactly one field whose values become the column headers.
#FieldRows			:	STRING. Specifies exactly one field whose values become the row headers.
#InfoTypes			:	LIST of STRINGS. Specifies the fields that are obtainable using StatsOfNumericalMessageField().
#						These values typically are 'Avg', 'Median', 'Std', 'Min', 'Max'
#=======================================================================================================
def HTMLTable( Filename, Data, InterestingFields, FieldColumn, FieldRows, InfoTypes = [ 'N', 'Avg', 'Median', 'Std', 'Max', 'Min' ] ):
	
	if type( InterestingFields ) is str:
		InterestingFields = [ InterestingFields ];
		
	if type( InterestingFields ) not in [ tuple, list ]:
		return 0;
	
	if type( Filename ) is not str:
		return 0;
	
	if type( FieldColumn ) is not str:
		return 0;
	
	if type( FieldRows ) is not str:
		return 0;
	
	if type( InfoTypes ) is str:
		InfoTypes = [ InfoTypes ];
	
	if type( InfoTypes ) not in [ tuple, list ]:
		return 0;
	
	Columns = UniqueStates( FieldColumn, Data );
	Rows = UniqueStates( FieldRows, Data );
				
	HTMLFile = open( Filename, 'w' );
	HTMLFile.write( "<head><title>"+str(InterestingFields)+"</title><META HTTP-EQUIV=REFRESH CONTENT=60></head>\n<body>\n" );
	
	GroupIndex = {};
	GroupedData = Select( [ FieldColumn, FieldRows], Data, [ FieldColumn, FieldRows] );
	for i in range( len( GroupedData ) ):
		Group = GroupedData[i];
		RowValue = str( UniqueStates( FieldRows, Group )[0] ).upper();
		ColValue = str( UniqueStates( FieldColumn, Group )[0] ).upper();
		GroupIndex[ ColValue+RowValue ] = i;
	
	if len( GroupedData )  == 0:
		HTMLFile.write( "<p><b>No Data!</b></p>" );
		HTMLFile.write( "</body>\n" );
		HTMLFile.close();
		return 0;
	
	for InterestingField in InterestingFields:
		HTMLFile.write( '<h1>' + InterestingField + '</h1>\n' );
		for FieldType in InfoTypes:
			HTMLFile.write( '<p>' + InterestingField + '\n' );
			HTMLFile.write( '<table>\n<tr><td BGCOLOR="#99990f">' + FieldType + '</td>' );
			for Column in Columns:
				HTMLFile.write( '<td BGCOLOR="#ccccff"><b>'+str(Column)+'</b></td>' );
			HTMLFile.write( '</tr>\n' );
			
			for Row in Rows:
				HTMLFile.write( '<tr><td BGCOLOR="#ff8080"><b>' + str(Row) + '</b></td>' );
				for Column in Columns:
					GroupPosition = GroupIndex.get( string.upper( str( Column ) + str( Row ) ), -1 );
					if GroupPosition >= 0:
						Group = GroupedData[GroupPosition];
						Msg = StatsOnNumericMessageField( InterestingField, Group );
						#Send( Msg )
						if Msg.has_key( FieldType ):
							HTMLFile.write( '<td>'+str( Msg[FieldType] ).replace("[","").replace("]","") + '</td>' );
						else:
							HTMLFile.write( '<td>n.a.</td>' );
					else:	
						HTMLFile.write( '<td>n.a.</td>' );
				HTMLFile.write( '</tr>\n' );
			HTMLFile.write( '</table></p>' );

	HTMLFile.write( "</body>\n" );
	HTMLFile.close();
	return 1;

#=======================================================================================================
#FUNCTION HTMLTable2( Filename, Data, InterestingFields, FieldColumn, FieldRows, InfoTypes )
#RETURNS 1/0 (SUCCESS / FAILED )
#This functions takes a list (or other list like objects like PythonFileReader) of messages
#and extract tabular information from the messages concerning >InterestingFields<. The
#number of rows and columns is automatically decided, depending how many different values
#are available in the database. This function only works for numerical types of >InterestingFields<.
#You can choose to generate statistical tables for average, median, standard deviation, maximum and minimum.
#
#Filename			:	STRING. Path to a new or overwritable html file
#Data				: 	LIST of DICTS. Database with messages.
#InterestingFields	:	STRING or LIST of STRINGS. Specifies which numerical field(s) you're interested in.
#FieldColumn		:	STRING. Specifies exactly one field whose values become the column headers.
#FieldRows			:	STRING. Specifies exactly one field whose values become the row headers.
#InfoTypes			:	LIST of STRINGS. Specifies the fields that are obtainable using StatsOfNumericalMessageField().
#						These values typically are 'Avg', 'Median', 'Std', 'Min', 'Max'
#=======================================================================================================
def HTMLTable2( Filename, Data, InterestingFields, FieldColumn, FieldRows, InfoTypes = [ 'N', 'Avg', 'Median', 'Std', 'Max', 'Min' ] ):
	
	if type( InterestingFields ) is str:
		InterestingFields = [ InterestingFields ];
		
	if type( InterestingFields ) not in [ tuple, list ]:
		return 0;
	
	if type( Filename ) is not str:
		return 0;
	
	if type( FieldColumn ) is not str:
		return 0;
	
	if type( FieldRows ) is not str:
		return 0;
	
	if type( InfoTypes ) is str:
		InfoTypes = [ InfoTypes ];
	
	if type( InfoTypes ) not in [ tuple, list ]:
		return 0;
	
	Columns = UniqueStates( FieldColumn, Data );
	Rows = UniqueStates( FieldRows, Data );
				
	HTMLFile = open( Filename, 'w' );
	HTMLFile.write( "<head><title>"+str(InterestingFields)+"</title><META HTTP-EQUIV=REFRESH CONTENT=60></head>\n<body>\n" );
	
	GroupIndex = {};
	GroupedData = Select( [ FieldColumn, FieldRows], Data, [ FieldColumn, FieldRows] );
	for i in range( len( GroupedData ) ):
		Group = GroupedData[i];
		RowValue = str( UniqueStates( FieldRows, Group )[0] ).upper();
		ColValue = str( UniqueStates( FieldColumn, Group )[0] ).upper();
		GroupIndex[ ColValue+RowValue ] = i;
	
	if len( GroupedData )  == 0:
		HTMLFile.write( "<p><b>No Data!</b></p>" );
		HTMLFile.write( "</body>\n" );
		HTMLFile.close();
		return 0;
	
	for InterestingField in InterestingFields:
		HTMLFile.write( '<h1>' + InterestingField + '</h1>\n' );
		
		for Row in Rows:
			HTMLFile.write( '<p>' + InterestingField + '\n' );
			HTMLFile.write( '<table>\n<tr><td BGCOLOR="#99990f">' + Row + '</td>' );
			for Column in Columns:
				HTMLFile.write( '<td BGCOLOR="#ccccff"><b>'+str(Column)+'</b></td>' );
			HTMLFile.write( '</tr>\n' );
			
			for FieldType in InfoTypes:	
				HTMLFile.write( '<tr><td BGCOLOR="#ff8080"><b>' + str(FieldType) + '</b></td>' );
				for Column in Columns:
					GroupPosition = GroupIndex.get( string.upper( str( Column ) + str( Row ) ), -1 );
					if GroupPosition >= 0:
						Group = GroupedData[GroupPosition];
						Msg = StatsOnNumericMessageField( InterestingField, Group );
						#Send( Msg )
						if Msg.has_key( FieldType ):
							HTMLFile.write( '<td>'+str( Msg[FieldType] ).replace("[","").replace("]","") + '</td>' );
						else:
							HTMLFile.write( '<td>n.a.</td>' );
					else:	
						HTMLFile.write( '<td>n.a.</td>' );
				HTMLFile.write( '</tr>\n' );
			HTMLFile.write( '</table></p>' );

	HTMLFile.write( "</body>\n" );
	HTMLFile.close();
	return 1;


#=======================================================================================================
#FUNCTION InRange()
#RETURNS SCALAR OR NDARRAY
#Checks array or scalars to be within specific ranges.
#Low		: FLOAT/INTEGER/ARRAY	Lower limits
#Data		: FLOAT/INTEGER/ARRAY	Data to be fit into range
#Low		: FLOAT/INTEGER/ARRAY	Upper limits
#=======================================================================================================
def InRange( Low, Data, High ):
	Result = maximum( Low, minimum( High, Data ) );
	return Result;




#=======================================================================================================
#FUNCTION IsNumeric()
#RETURNS 1/0
#Is it a numeric type of data?
#Anything			:	this can be some strange data
#=======================================================================================================
def IsNumeric( Anything ):
	if isscalar( Anything ):		
		if str( type( Anything ) ).upper().find( 'INT' ) + 1:
			return 1;
		if str( type( Anything ) ).upper().find( 'FLOAT' ) + 1:
			return 1;
	return 0;


#=======================================================================================================
#FUNCTION IsHashable()
#RETURNS 1/0
#Is it a hashable type of data?
#Anything			:	this can be some strange data
#=======================================================================================================
def IsHashable( Anything ):
	try:
		hash( Anything );
		return 1;
	except:
		return 0;


#=======================================================================================================
#FUNCTION IsRectMatrix()
#RETURNS 1/0
#Is it a rectangular matrix?
#Anything			:	this can be some strange data
#=======================================================================================================
def IsRectMatrix( MaybeMatrix ):
	if type( MaybeMatrix ) == ndarray:
		S = shape( MaybeMatrix );
		if len( S ) == 2:
			if S[0] == S[1]:
				return 1;
	return 0;


#=======================================================================================================
#FUNCTION IsVector()
#RETURNS 1/0
#Is it a numeric vector?
#Anything			:	this can be some strange data
#=======================================================================================================
def IsVector( MaybeVector ):
	if type( MaybeVector ) == ndarray:
		S = shape( MaybeVector );
		if len( S ) == 1 and S[0] > 0:
			return 1;
		if len( S ) == 2:
			if S[0] == 1 or S[1] == 1:
				return 1;
	return 0;




#=======================================================================================================
#FUNCTION MatchesWildcard()
#RETURNS 1/0	Yes/No
#WildExpression	:	STRING. Expression containing * and ? wildcards.
#String			:	STRING.	The string to be matched with WildExpression
#CaseSensitive	:	1/0.	Compares the strings case or non case sensitively.
#=======================================================================================================
def MatchesWildcard( WildExpression, String, CaseSensitive = 0 ):
	if not CaseSensitive:
		WildExpression 	= WildExpression.upper();
		String 			= String.upper();
	#end if CaseSensitive 
	
	Terms 		= WildExpression.replace( '?', ' ? ' ).replace('*', ' * ' ).replace( '\ ?', '?' ).replace( '\ *', '*' ).split();
	CurrentPos 	= 0;
	Strict 		= 1;
	for Term in Terms:
		if Term != '?' and Term != '*':
			Pos = String.find( Term, CurrentPos );
			if Strict:
				if not (Pos==CurrentPos):
					return 0;
				else: CurrentPos += len(Term);
			else:
				Strict = 1;
				if Pos < CurrentPos:
					return 0;
				else: CurrentPos = Pos + len(Term);
			#end if Strict
		elif Term == '?':
			CurrentPos += 1;
			if CurrentPos > len( String ):
				return 0;
		elif Term == '*':
			Strict = 0;
		#end if Term is ? or *
	return 1;
#end def MatchesWildcard


def MarginalizedDiagram( FieldOfInterest, Parameter, DataSet ):
	GraphMessage = {"Avg":None,
					"Std":None,
					"Max":None,
					"Min":None,
					"ParameterValues":[],
					"Parameter":Parameter,
					"Function":FieldOfInterest,
					"ParameterValues":[],
					"Status":"OK",
					};

	if type( FieldOfInterest ) is not str:
		GraphMessage["Status"] = "Failed! FieldOfInterest must be a string."
		return GraphMessage;

	if type( Parameter ) is not str:
		GraphMessage["Status"] = "Failed! Parameter must be a string."
		return GraphMessage;

	Data = Filter( [FieldOfInterest, Parameter], DataSet, [Parameter] );
	
	MergeList = [];
	for Group in Data: #out of scalars produce vectors
		Label = str(UniqueStates( Parameter, Group )[0]);
		GraphMessage["ParameterValues"] += [Label];
		NumericalContent = CombineNumericMessageFields( FieldOfInterest, Group )["Data"];
		MergeList += [ { Parameter:Label, "Merged":NumericalContent.T[0] } ];
		
	#print "DEBUG MergeList:"
	#print MergeList
	TotalNumericalValues = CombineNumericMessageFields( "Merged", MergeList )["Data"]; #out of vectors produce matrix
	#print TotalNumericalValues
	
	GraphMessage["Avg"] = TotalNumericalValues.mean( 1 );
	GraphMessage["Std"] = TotalNumericalValues.std( 1 );
	GraphMessage["Max"] = TotalNumericalValues.max( 1 );
	GraphMessage["Min"] = TotalNumericalValues.min( 1 );
	return GraphMessage;

#=======================================================================================================
#FUNCTION MergeMessages
#RETURNS Dictionary (PaREn Message)
#Merges messages that are totally different. Description fields are merged. Status content is ignored.
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def MergeMessages( Messages ):
	Result = { 'Description': '' };									#Description is the only field that is concatenated
	for Message in Messages:										#we iterate over all the messages
		Keys = Message.keys();										#and we iterate over all the attribute fields
		for Key in Keys:
			if Key.upper() == 'DESCRIPTION':						#if we find the exceptional Description field, ...
				Result['Description'] = Result['Description'] + ' ' + Message[Key]; #...then deal with it separately...
				continue;
			
			if Key.upper() == 'STATUS':								#Status messages are removed, because there is no reasonable way
				continue;											#to merge them (also not by overwriting)
			
																	#Otherwise just write/overwrite the field in the resulting message:
			Result[Key] = Message[Key];								#Be aware that content might get overwritten this way!
		#end for Key
	#end for Message
	return Result;													#return resulting message.
#end def MergeMessages()


#=======================================================================================================
#FUNCTION Messages2CSVTable( Filename, Messages, RedirectFields )
#RETURNS VOID
#This function converts the typical PaREn databases consisting of a list of dicts into a csv file.
#This function does not require the messages to have common fields. This way this command is
#robust against missing values. Of course you can obuse it in order to get totally crazy joins.
#Some fields are not atomic, but are pretty large lists and arrays like some training history data.
#You can specify those fields to be redirected into a separate csv file. In this case the original
#csv contains only the filename to the content.
#Filename		: STRING. Path to a new csv file.
#Messages		: LIST of DICTS. Can also be something like PythonFileReader
#RedirectFields	: LIST of STRINGS or one STRING. Contains the fields to be relayed to an own csv file.
#=======================================================================================================
def Messages2CSVTable( Filename, Messages, RedirectFields = [] ):
	#support single string scenarios for convenience
	if type( RedirectFields ) is str:
		RedirectFields = [RedirectFields];
	
	F2P, P2F = FieldIndex( Messages );
	FieldCount = len( F2P );
	if FieldCount == 0:
		return;
	
	Content = []; #this is a memory variable, that remember which content should go where. It will be given to FormatFreeCSVTable()
	
	#Get table headers!
	Row = NoneField( FieldCount );
	for i in range( FieldCount ):
		Row[i] = P2F[i];
	Content += [Row];

	#Arrange field in >Content<
	Line = 0;
	for Message in Messages:
		Line += 1;
		Row = NoneField( FieldCount );
		for i in range( FieldCount ):
			if P2F[i] in RedirectFields:
				SubData = Message.get( P2F[i], None );
				if SubData is not None:
					SubFile = Filename + P2F[i] + str(Line) + ".scv";
					FormatFreeCSVTable( SubFile, SubData ); 
					Row[i] = SubFile;
			else:
				Row[i] = Message.get( P2F[i], None );
				if Row[i] is None and P2F[i] == "ValidationError":
					print P2F[i]
					print i
					print Message
			#end if field is available in message
		Content += [Row];			
		#end for every column i
	#end for every row (message)
	
	#write content to csv file
	FormatFreeCSVTable( Filename, Content );
#end def


#=======================================================================================================
#FUNCTION Messages2Lists( Messages )
#RETURNS VOID
#This function converts the typical PaREn databases consisting of a list of dicts into a table made of lists.
#This function does not require the messages to have common fields. This way this command is
#robust against missing values. Of course you can abuse it in order to get totally crazy joins.
#Messages		: LIST of DICTS. Can also be something like PythonFileReader
#=======================================================================================================
def Messages2Lists( Messages ):
	F2P, P2F = FieldIndex( Messages );
	FieldCount = len( F2P );
	if FieldCount == 0:
		return;
	
	Content = []; #this is a memory variable, that remember which content should go where. It will be given to FormatFreeCSVTable()
	
	#Get table headers!
	Row = NoneField( FieldCount );
	for i in range( FieldCount ):
		Row[i] = P2F[i];
	Content += [Row];

	#Arrange field in >Content<
	Line = 0;
	for Message in Messages:
		Line += 1;
		Row = NoneField( FieldCount );
		for i in range( FieldCount ):
			Row[i] = Message.get( P2F[i], None );
		Content += [Row];			
		#end for every column i
	#end for every row (message)

	return Content;
#end def




#=======================================================================================================
#FUNCTION Messages2CSVTable( Filename, Messages, RedirectFields )
#RETURNS VOID
#This function converts the typical PaREn databases consisting of a list of dicts into a csv file.
#This function does not require the messages to have common fields. This way this command is
#robust against missing values. Of course you can obuse it in order to get totally crazy joins.
#Some fields are not atomic, but are pretty large lists and arrays like some training history data.
#You can specify those fields to be redirected into a separate csv file. In this case the original
#csv contains only the filename to the content.
#Filename		: STRING. Path to a new csv file.
#Messages		: LIST of DICTS. Can also be something like PythonFileReader
#RedirectFields	: LIST of STRINGS or one STRING. Contains the fields to be relayed to an own csv file.
#=======================================================================================================
def Messages2R( Filename, Messages, RedirectFields = [] ):
	#support single string scenarios for convenience
	if type( RedirectFields ) is str:
		RedirectFields = [RedirectFields];
	
	F2P, P2F = FieldIndex( Messages );
	FieldCount = len( F2P );
	if FieldCount == 0:
		return;
	
	Content = []; #this is a memory variable, that remember which content should go where. It will be given to FormatFreeCSVTable()
	
	#Get table headers!
	Row = NoneField( FieldCount );
	for i in range( FieldCount ):
		Row[i] = P2F[i];
	Content += [Row];

	#Arrange field in >Content<
	Line = 0;
	for Message in Messages:
		Line += 1;
		Row = NoneField( FieldCount );
		for i in range( FieldCount ):
			if P2F[i] in RedirectFields:
				SubData = Message.get( P2F[i], None );
				if SubData is not None:
					SubFile = Filename + P2F[i] + str(Line) + ".scv";
					RScriptExport( SubFile, SubData ); 
					Row[i] = SubFile;
			else:
				Row[i] = Message.get( P2F[i], None );
				if Row[i] is None and P2F[i] == "ValidationError":
					print P2F[i]
					print i
					print Message
			#end if field is available in message
		Content += [Row];			
		#end for every column i
	#end for every row (message)
	
	#write content to R script
	RScriptExport( Filename, Content );
#end def


#=======================================================================================================
#FUNCTION NewMessage()
#RETURNS DICTIONARY
#Creates a new message
#Description		:	STRING. something you'd like to put int....
#=======================================================================================================
def NewMessage( Description ):
	return { "Description":Description };



#=======================================================================================================
#FUNCTION NoneField
#RETURNS Field of Nones
#Creates an empty field of length n of nones
#=======================================================================================================
def NoneField( n ):
	Result = [];
	for i in range( n ):
		Result += [None];
	return Result;


#=======================================================================================================
#FUNCTION SortRows()
#RETURNS 1D/2D NDARRAY
#Statistical normalization.
#data			:	1D/2D NDARRAY or LIST. Data where rows are vectors or list.
#=======================================================================================================
def Normalize( data ):
	if len( shape( data ) ) == 1:
		O = mean( data );
		S = std( data );
		return (data-O)/S;
	
	if len( shape( data ) ) == 2:
		O = mean( data, 0 );
		S = std( data, 0 );
		return (data-O)/S;
	return array([]);


#=======================================================================================================
#FUNCTION PrintMessage
#RETURNS VOID
#Pretty print of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#PadLines	:	Integer. Prints PadLines empty lines after the message
#=======================================================================================================
def PrintMessage( Messages, Loud = 1, PadLines = 0, FullView = 0 ):
	if type( Messages ) == dict:
		Messages = [Messages];

	if type( Messages ) == list or type( Messages ) == tuple:
		for Message in Messages:
			TerminalMessage( Message, Loud, PadLines, FullView );
		#end for
	#end if
	return;
#end def 



#=======================================================================================================
#FUNCTION PrintHTML
#RETURNS VOID
#Pretty print of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#Filename	:   STRING. 	HTML file to be written into
#Refresh	:	INTEGER.	Number of seconds, before html is reloaded
#PadLines	:	Integer. 	Prints PadLines empty lines after the message
#FullView	:	1/0.		Abbreviate lists and arrays?
#=======================================================================================================
def PrintHTML( Messages, Filename, Title='PaREn', Refresh = 1000, PadLines = 0, FullView = 0 ):
	if type( Messages ) == dict:
		Messages = [Messages];

	if type( Messages ) == list or type( Messages ) == tuple:
		HTMLFile = open( Filename, 'w' );
			
		HTMLFile.write( "<head><title>"+str(Title)+"</title><META HTTP-EQUIV=REFRESH CONTENT=" + str( Refresh ) + "></head>\n<body><div align=center><h1>"+str(Title)+"</h1></div><hr />\n" );
		for Message in Messages:
			HTMLFile.write( HTMLMessage( Message, 1, PadLines, FullView ) );
		#end for
		HTMLFile.write( "</body>" );
	#end if
	return;
#end def 




#=======================================================================================================
#FUNCTION RandomName
#RETURNS STRING
#Returns a string with capital latin letters. 
#CodeLength		:	Integer. Length of the string.
#=======================================================================================================
def RandomName( CodeLength = 32 ):
	CodeName 	= int8(( rand( CodeLength ) * 25 ) + 65);			#trivia stuff. But we rely on ASCII coding!

	Result = '';
	for i in range( CodeLength ):
		Result += chr( CodeName[i] );								#connect the letters to form a string.
	#end for
	return Result;



#=======================================================================================================
#FUNCTION RenameFields()
#RETURNS FLOAT, STRING
#This function renames specified message fields to one common name.
#FieldNameList	:	LIST/TUPLE.	List of string representing fieldnames. Fieldnames can be augmented with * and ? wildcards. Use \* or \? in order to specify regular name elements.
#NewName		:	STRING.	Name of the field.
#Messages		:	LIST. List of PaREn Messages that shall be renamed.
#CaseSensitive	:	1/0. If set to 1 then name comaparisons are done in a case sensitive manner.
#=======================================================================================================
def RenameFields( FieldNamesList, NewName, Messages, CaseSensitive = 0 ):
	#SECTION A: INITIALIZATION
	Comments 	= 'OK';
	Code		= 1;
	
	#SECTION A: GUARD
	if type( FieldNamesList ) == str:
		FieldNamesList = [FieldNamesList];
	
	if type( FieldNamesList ) != list and type( FieldNamesList ) != tuple:
		return 0, "Wrong type of data for FieldNamesList!";

	if type( NewName ) != str:
		return 0, "Wrong type of data for NewName: Must be a string!";

	#SECTION B: RENAMING
	for Message in Messages:
		Keys = Message.keys();
		for OldName in FieldNamesList:
			for Key in Keys:
				if MatchesWildcard( OldName, Key, CaseSensitive ):
					if Message.has_key( NewName ):
						Comments = "Merging fields. Probably lost of data!";
						Code	 = 0.5;
					Message[NewName] = Message[Key];
					Message.pop( Key );
				#end if MatchesWildcard
			#end for all keys
		#end for all old names
	#end for all messages
	return Code, Comments
#end def


def RScriptExport( Filename, Data, FirstLineIsHeader=1, CaseName=RandomName( 4 ) ):
	Stream = open( Filename, "w" );

	RefVectors = Data;
	Vecs = len( RefVectors );
	Dims = 0;
	for Row in Data:
		Dims = max( Dims, len( Row ) );
		
	#print "DEBUG Dims=", Dims, "Vecs=",Vecs
	Features = [];
	
	StartRow = 0;
	if FirstLineIsHeader and Vecs >= 1:
		for Item in Data[0]:
			StartRow = 1;
			FeatName = AssureStringCompatibility( str( Item ), "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_", "X" )
			if FeatName[0] in "0123456789":
				FeatName = "dim" + FeatName;

			#print "DEBUG FeatName = ", FeatName
			Features += [ FeatName ]; #make sure there are resonable names
	else:
		for d in xrange( Dims ):
			Features += [ 'dim' + str( d ) ]; #otherwise give numeric names
	
	#in R you must specify your data by column. There is no way to specify it by row.
	#so here we go and extract columns of data:
	for d in xrange( Dims ): 
		Stream.write( Features[d] + ' <- c( ' );
		for v in xrange( StartRow, Vecs ): 
			try:				
				if RefVectors[v][d] is not None:
					if type( RefVectors[v][d] ) is str:
						Value = "'" + RefVectors[v][d] + "'";
					else:
						Value = str( RefVectors[v][d] );
				else: Value = "NA";
			except:
				Value = "NA";
			Stream.write( Value );
			if v < Vecs - 1:
				Stream.write( ", " );
		Stream.write( " );\n" );
	
	Variable = "Vectors" + CaseName;
	Stream.write( Variable + " <- data.frame( " );
	for d in xrange( Dims ):
		Stream.write( Features[d] );
		if d < Dims - 1:
			Stream.write( ", " );
	Stream.write( " );\n" );
	Stream.close();
	return Variable;
	#end def Export to R

#=======================================================================================================
#FUNCTION Select()
#RETURNS LIST of LISTS (2D LIST) of DICTIONARIES (PaREn Messages)
#FIELDS			: 	LIST OF STRINGS/STRING.	Fieldname(s) - case sensitive
#FROM			:	LIST OF DICTS. 			List of messages to scan
#GROUPBY		:	LIST OF STRINGS/STRING.	Fieldnames(s) to group by the results. - case sensitive
#WHERE			:	STRING.					Evaluable python string, where message attributes are available as variable names. Functions from sys, os and time modules can be used here. This can be a security risk!
#											(WHERE clause is a security risk! This is because statements to the system can be made using os.system() function.)
#
#USAGE EXAMPLE
#Messages	= [ {...}, {...}, {...} ];
#Selection	= Select( 'SomeField', Messages ); 						-> Returns only messages that have the field SomeField
#Selection	= Select( [ 'SomeField', 'OtherField' ], Messages ); 	-> Returns only messages that have the fields SomeField and OtherField
#Selection 	= Select( 'SomeField', Messages, WHERE='SomeField>2 and SomeField<5' );
#																	-> Returns only messages that have the field SomeField and this field has numeric values > 2 and < 5
#Selection	= Select( [ 'A', 'B' ], Messages, 'C' );				-> Messages with fields A,B and C where lists are grouped by values of C
#Selection	= Select( [ 'A', 'B' ], Messages, [ 'C', 'D' ], 'C+D/(A+int(B))<0.9' );
#																	-> returns messages that have fields A,B,C and D and that meet the condition 
#																		C+D/(A+int(B))<0.9 where B is a string type that must be converted to int.
#Selection 	= Select( FROM=Messages, GROUPBY='SomeField' );			-> returns messages that are grouped by the content of SomeField
#Selection 	= Select( FROM=Messages, GROUPBY='SomeField', WHERE='OtherField==1' );
#																	-> returns messages that are grouped by the content of SomeField. Note, that
#																	   all fields can be used for evaluation in the WHERE clause. When those
#																	   fields do not exist, the statement returns FALSE.
#ErroMessages = Select( 'Status', PaREnProtocol.ReportsOfModules(), WHERE='Status.upper()=="ERROR" and TIMESTAMP>time.time()-3600' );
#																	-> this would get you all error messages that occured within the last hour
#=======================================================================================================
def Select( FIELDS = [] , FROM = [], GROUPBY = [], WHERE=None ):
	#SECTION A: GUARD
	if type( FIELDS ) == str:
		FIELDS = [FIELDS];

	if type( GROUPBY ) == str:
		GROUPBY = [GROUPBY];

	if type( FIELDS ) != list and type( FIELDS ) != tuple:
		return [];
	
	#if type( FROM ) != list and type( FROM ) != tuple:
	#	return [];
	try:
		FROM.__getitem__
	except:
		return [];

	if type( GROUPBY ) != list and type( GROUPBY ) != tuple:
		return [];

	if type( WHERE ) != str and WHERE is not None:
		WHERE = None;

	#SECTION B: SELECTING MESSAGES
	NeccessaryKeys 	= FIELDS + GROUPBY;
	Selected 		= [];
	
	for Message in FROM:
		Satisfied = 1;
		for Key in NeccessaryKeys:
			if not Message.has_key( Key ):
				Satisfied = 0;
				break;
			#end def
		#end for
		
		if Satisfied:
			if WHERE is None:
				Selected += [Message];
			elif _HelperSelectWhere( Message, WHERE ):
				Selected += [Message];				
		#end if
	#end for

	if len( GROUPBY ) == 0:
		return [Selected];
	
	#SECTION C: PERFORM GROUPING
	Groups = [];
	for Message in Selected:
		AddAsNew = 1;
		for Group in Groups:
			CmpMsg 			= Group[0];
			IsTheSameThing 	= 1;

			for Key in GROUPBY:
				if Message[Key] != CmpMsg[Key]:
					IsTheSameThing = 0;
					break;
				#end if
			#end for

			if IsTheSameThing:
				AddAsNew = 0;
				Group += [Message];
				break;
			#end if
		#end for
		
		if AddAsNew:
			Groups += [[Message]];
		#end if
	#end for
	return Groups;
#end def Select()



#=======================================================================================================
#FUNCTION SetDescription()
#RETURNS VOID
#Uses default PaREn Description field to store a string information about the message
#Messages			:	DICT./LIST OF DICT's. Messages to set the field for.
#DescriptionString	:	STRING. the text
#=======================================================================================================
def SetDescription( Messages, DescriptionString ):
	#SECTION A: GUARDS
	if type( Messages ) != list and type( Messages ) != tuple:
		Messages = [Messages];

	if type( DescriptionString ) != str:
		if ModuleVerbosity:
			print "PaREnUtility::SetDescription(): Error! Description must be a string!";
		return;
	#end if

	#SECTION B: SET DESCRIPTIONS
	for Message in Messages:
		if type( Message ) != dict:
			if ModuleVerbosity:
				print "PaREnUtility::SetDescription(): Error! Message must be a dictionary!";				
			continue;
		#end if
	
		Message['Description'] = DescriptionString;
	#end for
#end def	




#=======================================================================================================
#FUNCTION SetStatus()
#RETURNS VOID
#Uses default PaREn Description field to store a string information about the message
#Messages			:	DICT./LIST OF DICT's. Messages to set the field for.
#StatValue			:	STRING/0..100/0.0..1.0. The status/progress information
#=======================================================================================================
def SetStatus( Messages, StatValue ):
	#SECTION A: GUARDS
	if type( Messages ) != list and type( Messages ) != tuple:
		Messages = [Messages];

	if type( StatValue ) != str and not isscalar( StatValue ):
		if ModuleVerbosity:
			print "PaREnUtility::SetStatus(): Error! Description must be a string, int{0..100} or float{0.0..1.0}!";
		return;
	#end if

	#SECTION B: SET DESCRIPTIONS
	for Message in Messages:
		if type( Message ) != dict:
			if ModuleVerbosity:
				print "PaREnUtility::SetStatus(): Error! Message must be a dictionary!";				
			continue;
		#end if
	
		Message['Status'] = StatValue;
	#end for
#end def	


#=======================================================================================================
#FUNCTION SortRows()
#RETURNS 2D NDARRAY
#Sorts rows of data by a given column.
#data			:	2D NDARRAY. data where rows are vectors.
#row			:	INT. Column to sort by.
#=======================================================================================================
def SortRows( data, row ):
	Data = array( data );
	Data = Data[Data[:,row].argsort()];
	return Data;


#=======================================================================================================
#FUNCTION StatsOnNumericMessageField
#RETURNS Dictionary
#Calcualtes Min, Max, Sum, Mean, Median, Average on a given numerical Field from the messages supplied
#Fields can be either scalar, 1D nd arrays or 1D ndlists
#Fieldname		:	STRING. Exact name of field. This function does not accept lists for Fieldname!!!
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def StatsOnNumericMessageField( Fieldname, Messages ):
	#SECTION A: INITIALIZATION
	Result	= { 'Description':'Statistical properties of "' + Fieldname + '" retrieved from ' + str( len( Messages ) ) + ' examples.', 'Status':'OK' };
	Data 	= [];
	N		= 0;

	#SECTION B: SEARCH FOR MINIMUM WIDTH OF DATA
	Width 	= inf;
	for Message in Messages:
		Entry = Message.get( Fieldname, None );
		if Entry is None:
			continue;
		if isscalar( Entry ):
			Width = 1;
		elif len( Entry ) < Width:
			Width = len( Entry );
		#end if
	#end for

	#SECTION C: PUT DATA INTO 2D LISTS
	for Message in Messages:
		Entry = Message.get( Fieldname, None );
		if Entry is None:
			continue;
		if isscalar( Entry ):
			Data += [[Entry]];
		elif Width < inf:
			Data += [Entry[:Width]];
		N += 1;
	#end for
	
	#SECTION D: TRY TO CONVERT TO 2D ARRAY
	try:
		Data			= FixPhysicalProperties( array( Data ), 2, 'PaREnUtility::StatsOnNumericMessageField()', 1 );
	except:
		Result['Status'] = 'Data is inconsistent in the messages!';
		return Result;
	#end try
	
	#SECTION E: PACKAGE RESULTS AND QUIT
	Result['N']			= N;
	Result['Min'] 		= Data.min(  0 );
	Result['Max']		= Data.max(  0 );
	Result['Avg']		= Data.mean( 0 );
	Result['Std']		= Data.std(  0 );
	Result['Median']	= median( Data );
	Result['Sum']		= Data.sum(  0 );
	Result['Cov']		= cov( transpose( Data ) );
	return Result;
#end def


#=======================================================================================================
#FUNCTION Table()
#RETURNS VOID
#Prints a table on the screen. The data is supplied as a message list and the content is then inferred.
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def Table( Messages, Headers=1, MaxW=30 ):
	T = Messages2Lists( Messages );
	TerminalTable( T, Headers, MaxW );


#=======================================================================================================
#FUNCTION TerminalInteraction()
#RETURNS String
#The message is displayed and a prompt is presented. The prompt must contain only the keywords
#that the user can choose from. Prompting information can be supplied with the message where it
#will override the function parameter. The user can type in the choice he wants
#Messages		:	LIST. List of dictionaries.
#=======================================================================================================
def TerminalInteraction( Message, Prompt = "OK", Commands = {} ):
	#SECTION A:	INITIALIZATION AND MESSAGE DECODING__________________________________________________________________________________
	Response 	= None;															#We have no input from the user yet
	FullFields	= 0;
	Prompt 		= Message.get( 'PROMPT', Prompt );								#Override Prompt with content from message
	Commands	= Message.get( 'COMMANDS', Commands );							#Override ResponseCommands with content from message
	PrintMessage( Message, FullView=FullFields );								#First of all display the content of the message
	
	Keys 		= Commands.keys();												#Get the possible actions, that were supplied...
	for Key in Keys:															#THIS IS DONE BECAUSE COMMANDS DOES NOT CONTAIN UPPERCASE LETTER NECCESSARILY!!!
		Commands[Key.upper()] = Commands[Key];									#...and associate the responses to actions
	
	PossibleAnswers = { 'EDIT':1, 'ABORT':1, 'HELP':1, 'ADD':1, 
						'DEL':1, 'TOGGLE':1, 'DRAW':1 };						#There are some defaults that are not displaye, but always work
	Alternatives 	= Prompt.upper().split();									#Separate the prompt into individual choices with keeping lettercase
	for Answer in Alternatives:
		PossibleAnswers[Answer] = 1;											#PossibleAnswers is used to quickly validate user input for validity!
	
	#SECTION B: USER INPUT LOOP______________________________________________________________________________________________________
	while not Response:
		#SECTION B.1: PROMPT USER FOR INPUT
		sys.stdout.write( Prompt + "? " );										#Display the choices for the user		
																				#read in the text line from user:
		Responses 	= sys.stdin.readline().upper().replace( '.', ' ').replace( ',', ' ').replace( '!', ' ' ).replace('?', ' ').replace( '\'', ' ' ).split();						
		for Response in Responses:												#the user might have typed in a lot of stuff
																				#assume this is the keyword you can understand
			if PossibleAnswers.get( Response, 0 ):								#therefore it must be able to find it in PossibleAnswers
				break;															#if it is possible, then keep on going
			#end if																#if it is not possible, then try next word
		#end for																#if all words were bad, than continue with last word
		print "PaREn: I understand '" , Response , "'";							#give user visual response
			
		#SECTION B.2: EVALUATE USER RESPONSE
		if PossibleAnswers.get( Response, 0 ):									#Is it valid user response?
			if Response == "HELP":												#Ok, user wants some help
				HelpMessage = { "Description":'The answers are not case sensitive. The following commands can be used on the prompt always:' };
				HelpMessage['Help'] 	= 'Takes no parameters and shows you the interactive terminal help.';
				HelpMessage['Edit']		= 'Takes no parameters. Menu driven editing.';
				HelpMessage['Abort'] 	= 'Takes no parameters. Aborts the interaction.';
				HelpMessage['Add']		= 'Takes no parameters. Adds a new field. Use edit to change its value.';
				HelpMessage['Del']		= 'Takes no parameters. Removes a specified field.';
				HelpMessage['Toggle'] 	= 'Takes no parameters. Extends/abbreviates long fields.';
				HelpMessage['Draw']		= 'Takes no parameters. Menu driven selection of data to draw';
				PrintMessage( HelpMessage );
				Response = None;
				continue;
			#end if

			if Response == "ADD":
				try:
					sys.stdout.write( "Name: " );
					Name = str( sys.stdin.readline().strip() );
					sys.stdout.write( "Value: " );
					Value = eval( sys.stdin.readline() );
					Message[Name] = Value;
					PrintMessage( Message, FullView=FullFields  );
				except:
					print "You are making some mistake but I don't know what.";
				Response = None;
				continue;
			#end if

			if Response == "DEL":
				try:
					sys.stdout.write( "Name: " );
					Name = str( sys.stdin.readline().strip() );
					Message.pop( Name );
					PrintMessage( Message, FullView=FullFields  );
				except:
					print "Can't do! Field probably doesn't exist. Here you must be case sensitive!";
				Response = None;
				continue;
			#end if

			if Response == "TOGGLE":
				FullFields = 1 - FullFields;
				PrintMessage( Message, FullView=FullFields  );
				Response = None;
				continue;
			#end if

			if Response == "EDIT":												#Ok, user want to edit the message before respondig
				#SECTION B.2.1: INITIALIZE EDIT MENU AND PRINT IT
				print "Please select the field to edit:";
				Interrogation = {};												#This dictionary save NUMBERS->ATTRIBUTENAMES relationships
				Keys = [ 'Cancel' ] + Message.keys();							#Insert a default position cancel where nothing is changed
				for Key in Keys:												#^
					if Key[0] == '_':											#|
						continue;												#|Build up the Interrogation dictionary for quick
					print "\t(", len(Interrogation), ")\t", Key					#|decoding of user input to attribute name
					Interrogation[len(Interrogation)] = Key;					#|
				#end for														#v
				
				#SECTION B.2.2: READ IN EDIT MENU
				Choice = None;													#User has made no input yet
				while Choice is None:											#while user is reluctant to give you something reasonable
					try:														#we just try, because users are dangerous
						Choice = int( sys.stdin.readline() );					#oh, we want an integer...
						if Choice < 0 or Choice >= len( Interrogation ):		#...in a special range
							Choice = None;										#every other number confuses us
							print "You must type in a number from the menu!";						
					except:
						print "You must type in a number from the menu!";		#ups, I knew that users are dangerous people!
						Choice = None;
				#end while
				
				#SECTION B.2.3: DEAL WITH EDITING REQUEST FOR ATTRIBUTE
				SelectedField = Interrogation[Choice];							#OK, we must deal with this field now
				if SelectedField != 'Cancel':									#unless it is cancel, obviously
					if type( Message[SelectedField] ) == dict:					#take care of this separately!
						TerminalInteraction( Message[SelectedField], 'R (Return)' );
					else:
						#SECTION B.2.3.1: PROMPT THE ATTRIBUTE EDIT MENU:
						print "Please enter new content for the field ", SelectedField
						OldContent = str( Message[SelectedField] );					#Retrieve content of attribute
						if len(OldContent) > 100:									#And don't present all of it
							OldContent = OldConten[:100];							#it's enough to get an idea of it
							
						print "Current content is '", OldContent, "'"				#present content excerpt
						TypeInfo = Message.get( '_'+SelectedField, None );			#get possible input alternatives
						if TypeInfo is not None:									#if type info is available
							if type( TypeInfo ) == list:							#deal with list separately
								print "Please select one of them: "
								i = 0;												#enumerative index for the list
								print "\t[-1] Cancel"
								for Possibility in TypeInfo:						#print choices to pick from the list
									print "\t[", i, "]\t", Possibility
									i += 1;
								#end for
							else:													#if it's not a list then just print the type:
								print "Expected field type is: ", TypeInfo
						#end if
						sys.stdout.write( ":" );									#prompt the user
											
						#SECTION B.2.3.2: READ IN USER INPUT FOR ATTRIBUTE
						try:														#As everybody knows: users are evil
							if TypeInfo is None:
								NewContent = eval( sys.stdin.readline() );			#we had no type info, so read in as any type of expression
							elif type( TypeInfo ) == list:
								NewContent = TypeInfo[ int( eval( sys.stdin.readline() ) ) ];	#if list exception readin as integer choice
							else:
								NewContent = TypeInfo( eval( sys.stdin.readline() ) ); #or just convert to the predefined type
							#end if
							Message[SelectedField] = NewContent;					#and save new attribute value
						except:
							print "Something is wrong with the content. Try again, maybe?"					
						#end try
					#end if element is a message
				Response = None;												#If you get here, your editing is finished, but not your session
				PrintMessage( Message, FullView=FullFields  );					#you still want to know what to do from the original topmost prompt
				continue;														#start over again
			
			if Response == "DRAW":												#Ok, user want to draw some fields...
				#SECTION B.2.1: INITIALIZE EDIT MENU AND PRINT IT
				print "Please select the field to edit:";
				Interrogation = {};												#This dictionary save NUMBERS->ATTRIBUTENAMES relationships
				AllKeys = Message.keys();										#Insert a default position cancel where nothing is changed
				Keys 	= [ 'Cancel' ];
				
				for Key in AllKeys:												#^
					if Key[0] == '_':											#|
						continue;												#|Build up the Interrogation dictionary for quick
					if type( Message[Key] ) == list:							#|
						Keys += [ Key ];										#|
																				#|
				for Key in Keys:												#|
					print "\t(", len(Interrogation), ")\t", Key					#|decoding of user input to attribute name
					Interrogation[len(Interrogation)] = Key;					#|
				#end for														#v
				
				#SECTION B.2.2: READ IN EDIT MENU
				Choice = None;													#User has made no input yet
				while Choice is None:											#while user is reluctant to give you something reasonable
					try:														#we just try, because users are dangerous
						Choice = int( sys.stdin.readline() );					#oh, we want an integer...
						if Choice < 0 or Choice >= len( Interrogation ):		#...in a special range
							Choice = None;										#every other number confuses us
							print "You must type in a number from the menu!";						
					except:
						print "You must type in a number from the menu!";		#ups, I knew that users are dangerous people!
						Choice = None;
				#end while
				
				#SECTION B.2.3: DEAL WITH DRAWING
				SelectedField = Interrogation[Choice];							#OK, we must deal with this field now
				if SelectedField != 'Cancel':									#unless it is cancel, obviously
					Data = Message[SelectedField];
					ion();
					figure();
					plot( Data );
					title( str( SelectedField ) );
				#end if
				
				Response = None;												#If you get here, your editing is finished, but not your session
				PrintMessage( Message, FullView=FullFields  );					#you still want to know what to do from the original topmost prompt
				continue;														#start over again

			
			Message["ANSWER"] = Response;										#BUT if your message was one of the choices in the AcceptedAnswers
			ToExecute = Commands.get( Response, None );							#Then you have to quit, but before you quit, you might have to execute some commands
			if ToExecute is not None:											#if there is something supplied for execution
				if type( ToExecute ) == str:									#and this something is a string
					ToDo = ToExecute.split( '\n' );								#then this string is probably made from python lines
					for Action in ToDo:											#for all lines...
						try:
							Message["ANSWER"] = eval( Action );					#...try the classical eval function. But it doesn't work for everything
						except:
							exec Action											#in some cases they exec function
						#end try
					#end for Action
				else:
					ToExecute();												#If ToExecute is not a string, assume it is a function pointer.
		else:
			Response = None;													#this happens here, when user bubbles nonsense....
			print "PaREn: You must choose from the alternatives or ask for help!"
		#end if		
	#end while	
	return Response;
#end def TerminalInteraction();




#=======================================================================================================
#FUNCTION TerminalMessage
#RETURNS VOID
#Pretty print of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#PadLines	:	Integer. Prints PadLines empty lines after the message
#=======================================================================================================
def TerminalMessage( Message, Loud = 1, PadLines = 0, FullView = 0 ):
	Out = TextMessage( Message, Loud, PadLines, FullView );
	print Out;
#end def TerminalMessage()


#=======================================================================================================
#FUNCTION Messages2Lists( Messages )
#RETURNS VOID
#This function converts the typical PaREn databases consisting of a list of dicts into a table made of lists.
#This function does not require the messages to have common fields. This way this command is
#robust against missing values. Of course you can abuse it in order to get totally crazy joins.
#Messages		: LIST of DICTS. Can also be something like PythonFileReader
#=======================================================================================================
def TerminalTable( Data, Headers=1, MaxColumnWidth=30 ):
	if Data is None:
		return; 
	
	Rows = len( Data );
	Columns = 0;
	for r in xrange( Rows ):
		Columns = max( Columns, len( Data[r] ) );
		
	ColumnWidths = ones( Columns );
	for r in xrange( Rows ):
		for c in xrange( Columns ):
			if Data[r] is None: continue;
			try:
				if c < len( Data[r] ):
					ColumnWidths[c] = min( max( 1, ColumnWidths[c], len( str( Data[r][c] ) ) ), MaxColumnWidth );
			except:
				pass;

	Horizontal = "";
	for c in xrange( Columns ):
		Horizontal += "+".ljust( ColumnWidths[c]+1, '-' );
	Horizontal += "+";
	print Horizontal
		
	for r in xrange( Rows ):
		if r==1 and Headers:
			print Horizontal;
			
		Out = "";
		for c in xrange( Columns ):
			if c < len( Data[r] ) and Data[r][c] is not None:
				Value = str( Data[r][c] );
				if len( Value ) > MaxColumnWidth:
					Value = Value[:MaxColumnWidth-3]+"...";
				Out += "|" + Value.center( int( ColumnWidths[c] ) ) 
			else:
				Out += "|" + "*".center( int( ColumnWidths[c] ) ) 
		Out += "|";
		print Out;
	
	print Horizontal
#end def TerminalTable


#=======================================================================================================
#FUNCTION TextMessage
#RETURNS STRING
#Pretty print of a dictionary with some conventions. (PaREn Message)
#									PaREn Messages are dictionaries 
#									with a specific set of conventions. The key "Description" is
#									displayed as header of the message and does not have a ShadowKey.
#									Messages can contain ShadowKeys that have the same name with an
#									underscore: "ValueX" -> "_ValueX"
#									ShadowKeys are displayed differently. ShadowKeys are typically lists
#									that enumerate valid content or contain types that can be checked
#									against with a isinstance() command.
#Message	:	Dictionary. PaREn Message.
#PadLines	:	Integer. Prints PadLines empty lines after the message
#=======================================================================================================
def TextMessage( Message, Loud = 1, PadLines = 0, FullView = 0 ):
	#SECTION A: GUARDS SECTION
	if not PrintingToTerminalAllowed:
		return;
	
	if not Loud:
		return;
	
	if not isinstance( Message, dict ):
		return;
	
	if len( Message.keys() ) == 0:
		return;
	
	#SECTION B:	Print Header
	Output = "";
	ShouldPrintDataBar = 0;
	Output += "+--PaREn-MESSAGE----------------------------------------------------------------------\n";
	if Message.get( "Description", None ) is not None:
		TextMessage = Message["Description"];									#The following stuff is just a complicated print.
		Cursor		= 0;														#It is done because some lines are just long, so we want to break them.
		Length		= len( TextMessage );
		while Length - Cursor > 82:												#Break as many times as the string to print is longer than the line limit.
			RightCursor = TextMessage.find( ' ', Cursor + 1 );					#We break at space characters.
			while RightCursor - Cursor < 82:									#Keep looking at the first occurence of a space that would not fit into...
				LastCursor	= RightCursor;										#...the line and take the previous one for breaking.
				RightCursor = TextMessage.find( ' ', RightCursor + 1 );
				if RightCursor<0:
					break;
			#end while
			Output += "| " + str( TextMessage[Cursor:LastCursor] ) + "\n";		#print the line portion that actually did fit into the line limit
			Cursor = LastCursor + 1;
		#end while
		Output += "| " + str( TextMessage[Cursor:Length] ) + "\n";				#print the last piece of the line.
		ShouldPrintDataBar = 1;
	#end if

	#SECTION C: FIND OPTIMAL LEFT BORDER WIDTH
	OptimalBorder = 6;
	Keys = Message.keys();														#Get all the keys we would like to present
	
	Remarks = '';
	for i in range( len(Keys)-1,-1,-1 ):
		if type( Keys[i] ) != str:			
			Remarks += str( Keys.pop(i) ) + ' ';
		#end if
	#end for
	
	for key in Keys:
		if key.upper() == "DESCRIPTION":										#This key has been printed already
			continue;

		if key.upper() == "STATUS":												#We will deal with this later.
			continue;

		if key[0] == '_':
			continue;
		
		if OptimalBorder < len( key ):
			OptimalBorder = len( key );
	#end for

	if len( Remarks ) > 0:
		Message = Message.copy();
		Message['Remarks'] = 'The following fields can not be displayed:' + Remarks;
		Keys += ['Remarks'];
	#end if

	#SECTION D: WRITE DATA SECTION
	HasPrintedDataBar 	= 0;
	StatusKey 			= None;
	LengthLimit			= 60;
	for key in Keys:
		if key.upper() == "DESCRIPTION":										#This key has been printed already
			continue;

		if key.upper() == "STATUS":												#We will deal with this later.
			StatusKey = key;
			continue;
		
		if key[0] == '_':														#If a key is not a ShadowKey then print it like that:
			continue;

		if ShouldPrintDataBar and not HasPrintedDataBar:
			Output += "|:::::::::::::::::::::::::::::::::::::::::DATA::::::::::::::::::::::::::::::::::::::::\n";
			HasPrintedDataBar = 1;

		Value = Message[key];													#Find the value of the key
		if Value is None:														#Sometimes Values are None to indicate a Default setting
			Value = "Default";
			
		if not FullView and type( Value ) == dict and len( str( Value ) ) > LengthLimit:
			Value = "(PaREn Message)";
		
		
		if not FullView and len( str( Value ) ) > LengthLimit:					#In this clause we shorten up the look of the strings
			NewValue = str(Value)[:LengthLimit] + '...';						#and add some extra infos in compensation
			if type( Value ) == list:
				NewValue += "\n" + "LIST of " + str( len( Value ) ) + " elements.";
			if type( Value ) == tuple:
				NewValue += "\n" + "TUPLE of " + str( len( Value ) ) + " entries.";
			if type( Value ) == ndarray:
				NewValue += "\n" + "ARRAY of " + str( shape( Value ) ) + " size.";
			Value = NewValue;
		
		if len( shape( Value ) ) > 1 and type( Value ) != str:					#this clause deals with 2d arrays that need some more space.
			Value = '\n' + str( Value ) + '\n';			
			Value = str(Value).replace( '\n', '\n|'+string.rjust( "", OptimalBorder + 6 ) );
		elif str( Value ).find( '\n' )+1:
			Value = str(Value).replace( '\n', '\n|'+string.rjust( "", OptimalBorder + 6 ) );			
		#end if
		
		if Message.get( '_' + key, None ) is not None:						#if it has as shadow key then print...
			ListMotivatedInsertion = "\t"
			Type = Message['_'+key]
			if type( Type ) == list or type( Type ) == tuple:
				ListMotivatedInsertion = " where possible are: ";

			Output += "| " + string.rjust( key, OptimalBorder ) + ' : ' + str( Value ) + ListMotivatedInsertion + str( Message['_'+key] ) + "\n"; #...it with the legal Range
		else:
			Output += "| " + string.rjust( key, OptimalBorder ) + ' : ' + str( Value ) + "\n";				#otherwise just print it.
		#end if
	#end for	
	
	#SECTION D: Print Status
	if Message.get( StatusKey, None ) is not None:
		Value = Message[StatusKey];
		Output += "|::::::::::::::::::::::::::::::::::::::::STATUS:::::::::::::::::::::::::::::::::::::::\n" ;
		if type( Value ) == str:
			if str(Value).upper().find( "OK" )+1 or str(Value).upper().find( "SUCCESS" )+1\
			or str(Value).upper().find( "ALRIGHT" )+1 or str(Value).upper().find( "GOOD" )+1:
				Output += "| " + string.center( '>> ' + Value + ' <<', 82 ) + "\n";			
			elif str(Value).upper().find( "WARNING" )+1:
				Output += "| " + string.center( '* * * ' + Value + ' * * *', 82 ) + "\n";		
			elif str(Value).upper().find( "ERROR" )+1 or str(Value).upper().find( "FAIL" )+1:
				Output += "| " + string.center( ':( ' + Value + ' :(', 82 ) + "\n";
			else:
				Output += "| " + string.center( Value, 82 ) + "\n";
			#end if
		#end if TYPE IS STRING
			
		if type( Value ) == type( 0.0 ):
			f = max( min( Value, 1.0 ), 0.0 );
			c = fix( 60 * f );
			Bars = "";
			for i in range( int( c ) ):
				Bars += "|";
			Output +=  "|" + string.center( "[" + string.ljust( Bars, 60 ) + "] " + str(fix(100*f)) + "%", 82 ) + "\n";
		#end if TYPE IS FLOAT

		if type( Value ) == type( 0 ):
			f = max( min( Value, 100 ), 0 );
			c = fix( f * 0.6 );
			Bars = "";
			for i in range( int( c ) ):
				Bars += "|";
			Output += "|" + string.center( "[" + string.ljust( Bars, 60 ) + "] " + str(f) + "%", 82 ) + "\n";
		#end if TYPE IS INT
	#end if
	Output += "+-------------------------------------------------------------------------------------\n";
	
	#SECTION E: Write Pad Lines
	for i in range( PadLines ):
		Output += "\n";
		
	return Output;
#end def TerminalMessage2()


#=======================================================================================================
#FUNCTION TypeOf
#RETURNS anything
#Returns a type information for given message and key
#=======================================================================================================
def TypeOf( Message, FieldName ):
	if not type( Message ) == dict:
		return type( None );
	
	if Message.has_key( '_' + FieldName ):
		return Message.get( '_' + FieldName, None );
	
	return type( Message.get( FieldName, None ) );
#end def



#=======================================================================================================
#FUNCTION UniqueMessages()
#RETURNS LIST OF DICTIONARIES
#Sorts out duplicates of messages
#ListOfMessages		:	LIST OF DICTIONARIES. The messages you would like to 'compact'
#=======================================================================================================
def UniqueMessages( ListOfMessages ):
	NewPile = [];
	for Check in range( len( ListOfMessages ) ):
		Passed = 1;
		for ToCompareTo in range( Check+1, len( ListOfMessages ) ):
			if ListOfMessages[Check] == ListOfMessages[ToCompareTo]:
				Passed = 0;
				break;
			#end if
		#end for

		if Passed:
			NewPile += [ ListOfMessages[Check] ];
	#end for
	return NewPile;
#end def



#=======================================================================================================
#FUNCTION UniqueStates()
#RETURNS LIST OF DICTIONARIES
#Find for a given field in a set of messages all values that occur for it.
#Field			:	STRING. Fieldname you're interested in
#Messages		:	LIST OF DICTIONARIES. The messages you would like to 'compact'
#=======================================================================================================
def UniqueStates( Field, Messages ):
	Values = {};
	for Message in Messages:
		try:
			Values[Message[Field]] = 1;
		except: pass;
	try:
		x = Values.keys();
		x.sort();
		return x;
	except:
		return Values.keys();


#=======================================================================================================
#FUNCTION VectorCount
#RETURNS INT
#Returns how many vectors are stored in the data
#data		: ndarray
#=======================================================================================================
def VectorCount( data ):
	if data.ndim >= 2:
		N = shape( data.T )[1]
	else:
		N = shape( data.T )[0]
	return N;


#=======================================================================================================
#FUNCTION Dimensionality
#RETURNS INT
#Returns how long the vectors are
#data		: ndarray
#=======================================================================================================
def Dimensionality( data ):
	if data.ndim >= 2:
		N = shape( data.T )[0]
	else:
		N = 1;
	return N;


#=======================================================================================================
#FUNCTION WelcomeMessage
#RETURNS DICTIONARY
#Returns a message with a funny welcome description
#=======================================================================================================
def WelcomeMessage():
	Result = { 'Description':'Welcome' };
	T		= time.localtime();
	Choice  = fix(rand()*4);

	if rand() < 0.3:
		return Result;

	TimeMessage = "";
	if T[3] > 4 and T[3] < 13:
		TimeMessage = "Good Morning!"
	if T[3] > 12 and T[3] < 19:
		TimeMessage = "Have fun at work!"
	if T[3] > 18 and T[3] < 4:
		TimeMessage = "Are you still at work?"


	if Choice == 0:
		Result['Description'] = "Hi! " + TimeMessage;
	if Choice == 1:
		Result['Description'] = "There you are! " + TimeMessage;
	if Choice == 2:
		Result['Description'] = ":) " + TimeMessage;
	if Choice == 3:
		Result['Description'] = "Here is PaREn! I'm the prototype of HAL9000. " + TimeMessage;
	return Result;
#end def
	

#=======================================================================================================
#FUNCTION _HelperSelectWhere()
#RETURNS LIST
#Executes a logical evaluation of the fields given a rule in string form (python code).
#Message		:	DICTIONARY. PaREn Message that was in TerminalInteraction()
#Rule			:	STRING. A string with a python expression that would evaluate to boolean or int
#=======================================================================================================
def _HelperSelectWhere( Message, Rule ):
	sys._getframe().f_locals.update( Message );
	try:
		Result = eval( Rule );
	except:
		Result = 0;
	return Result;


#=======================================================================================================
#FUNCTION WaitForUser()
#RETURNS VOID
#Waits for user press ENTER
#=======================================================================================================
def WaitForUser():
	print "PLEASE PRESS ENTER IN ORDER TO CONINUE"
	sys.stdin.readline()




def InteractiveSelfTest1():
	print "T0123: Testing input dla"
	Msg = NewMessage( "1. Try to add value 2. try to remove value 3. try to rename ..." );
	TerminalInteraction( Msg, "Take Leave", { "Take":"print 'OK'", "Leave":"print 'Sorry'" } );
	

#=======================================================================================================
#MODULE TESTS
#=======================================================================================================
#Here I  test ....
if type( WelcomeMessage() ) is not dict:
	raise Exception, "T0001: WelcomeMessage is not of the right type (dict)!"

#=======================================================================================================
#END OF FILE
#=======================================================================================================
